# pylint: disable=too-many-lines
from unittest import TestCase

from pcs.common.corosync_conf import CorosyncNodeAddressType
from pcs.common.reports import codes as report_codes
from pcs.lib.cib.node import PacemakerNode
from pcs.lib.corosync import (
    config_validators,
    constants,
    node,
)

from pcs_test.tools import fixture
from pcs_test.tools.assertions import assert_report_item_list_equal
from pcs_test.tools.custom_mock import patch_getaddrinfo

forbidden_characters_kwargs = dict(
    allowed_values=None,
    cannot_be_empty=False,
    forbidden_characters=r"{}\n\r",
)


class AddLink(TestCase):
    # pylint: disable=too-many-public-methods
    def setUp(self):
        self.new_addrs = {
            "node1": "addr1-new",
            "node2": "addr2-new",
            "node3": "addr3-new",
        }
        self.transport = constants.TRANSPORTS_KNET[0]
        self.existing_link_list = ["0", "1", "3"]
        self.coro_nodes = [
            node.CorosyncNode(
                f"node{i}",
                [
                    node.CorosyncNodeAddress(f"addr{i}-{j}", f"{j}")
                    for j in self.existing_link_list
                ],
                i,
            )
            for i in [1, 2, 3]
        ]
        self.pcmk_nodes = []
        patch_getaddrinfo(self, self.new_addrs.values())

    def test_success(self):
        assert_report_item_list_equal(
            config_validators.add_link(
                self.new_addrs,
                {
                    "linknumber": "2",
                    "link_priority": "2",
                    "mcastport": "5405",
                    "ping_interval": "100",
                    "ping_precision": "10",
                    "ping_timeout": "50",
                    "pong_count": "5",
                    "transport": "udp",
                },
                self.coro_nodes,
                self.pcmk_nodes,
                self.existing_link_list,
                self.transport,
                constants.IP_VERSION_64,
            ),
            [],
        )

    def test_success_no_options(self):
        assert_report_item_list_equal(
            config_validators.add_link(
                self.new_addrs,
                {},
                self.coro_nodes,
                self.pcmk_nodes,
                self.existing_link_list,
                self.transport,
                constants.IP_VERSION_64,
            ),
            [],
        )

    def _assert_bad_transport(self, transport):
        assert_report_item_list_equal(
            config_validators.add_link(
                self.new_addrs,
                {},
                self.coro_nodes,
                self.pcmk_nodes,
                self.existing_link_list,
                transport,
                constants.IP_VERSION_64,
            ),
            [
                fixture.error(
                    report_codes.COROSYNC_CANNOT_ADD_REMOVE_LINKS_BAD_TRANSPORT,
                    add_or_not_remove=True,
                    actual_transport=transport,
                    required_transports=["knet"],
                )
            ],
        )

    def test_transport_udp(self):
        self._assert_bad_transport("udp")

    def test_transport_udpu(self):
        self._assert_bad_transport("udpu")

    def test_ping_interval_without_ping_timeout(self):
        assert_report_item_list_equal(
            config_validators.add_link(
                self.new_addrs,
                {
                    "linknumber": "2",
                    "ping_interval": "100",
                },
                self.coro_nodes,
                self.pcmk_nodes,
                self.existing_link_list,
                self.transport,
                constants.IP_VERSION_64,
            ),
            [
                fixture.error(
                    report_codes.PREREQUISITE_OPTION_IS_MISSING,
                    option_name="ping_interval",
                    option_type="link",
                    prerequisite_name="ping_timeout",
                    prerequisite_type="link",
                ),
            ],
        )

    def test_ping_timeout_without_ping_interval(self):
        assert_report_item_list_equal(
            config_validators.add_link(
                self.new_addrs,
                {
                    "linknumber": "2",
                    "ping_timeout": "50",
                },
                self.coro_nodes,
                self.pcmk_nodes,
                self.existing_link_list,
                self.transport,
                constants.IP_VERSION_64,
            ),
            [
                fixture.error(
                    report_codes.PREREQUISITE_OPTION_IS_MISSING,
                    option_name="ping_timeout",
                    option_type="link",
                    prerequisite_name="ping_interval",
                    prerequisite_type="link",
                ),
            ],
        )

    def test_too_many_links(self):
        existing_link_list = [str(x) for x in range(constants.LINKS_KNET_MAX)]
        coro_nodes = [
            node.CorosyncNode(
                f"node{i}",
                [
                    node.CorosyncNodeAddress(f"addr{i}-{j}", f"{j}")
                    for j in existing_link_list
                ],
                i,
            )
            for i in [1, 2, 3]
        ]
        assert_report_item_list_equal(
            config_validators.add_link(
                self.new_addrs,
                {},
                coro_nodes,
                self.pcmk_nodes,
                existing_link_list,
                self.transport,
                constants.IP_VERSION_64,
            ),
            [
                fixture.error(
                    # pylint: disable=line-too-long
                    report_codes.COROSYNC_CANNOT_ADD_REMOVE_LINKS_TOO_MANY_FEW_LINKS,
                    links_change_count=1,
                    links_new_count=(constants.LINKS_KNET_MAX + 1),
                    links_limit_count=constants.LINKS_KNET_MAX,
                    add_or_not_remove=True,
                )
            ],
        )

    def test_add_last_allowed_link(self):
        existing_link_list = [
            str(x) for x in range(constants.LINKS_KNET_MAX - 1)
        ]
        coro_nodes = [
            node.CorosyncNode(
                f"node{i}",
                [
                    node.CorosyncNodeAddress(f"addr{i}-{j}", f"{j}")
                    for j in existing_link_list
                ],
                i,
            )
            for i in [1, 2, 3]
        ]
        assert_report_item_list_equal(
            config_validators.add_link(
                self.new_addrs,
                {},
                coro_nodes,
                self.pcmk_nodes,
                existing_link_list,
                self.transport,
                constants.IP_VERSION_64,
            ),
            [],
        )

    def test_link_already_exists(self):
        linknumber = self.existing_link_list[1]
        assert_report_item_list_equal(
            config_validators.add_link(
                self.new_addrs,
                {"linknumber": linknumber},
                self.coro_nodes,
                self.pcmk_nodes,
                self.existing_link_list,
                self.transport,
                constants.IP_VERSION_64,
            ),
            [
                fixture.error(
                    report_codes.COROSYNC_LINK_ALREADY_EXISTS_CANNOT_ADD,
                    link_number=linknumber,
                )
            ],
        )

    def test_missing_node_addrs(self):
        broken_nodes = sorted(self.new_addrs.keys())[1:2]
        for node_name in broken_nodes:
            del self.new_addrs[node_name]
        pcmk_nodes = [PacemakerNode("node-remote", "addr-remote")]
        assert_report_item_list_equal(
            config_validators.add_link(
                self.new_addrs,
                {},
                self.coro_nodes,
                pcmk_nodes,
                self.existing_link_list,
                self.transport,
                constants.IP_VERSION_64,
            ),
            [
                fixture.error(
                    report_codes.COROSYNC_BAD_NODE_ADDRESSES_COUNT,
                    actual_count=0,
                    min_count=1,
                    max_count=1,
                    node_name=node_name,
                    node_index=None,
                )
                for node_name in broken_nodes
            ],
        )

    def test_empty_node_addr(self):
        broken_nodes = sorted(self.new_addrs.keys())[1:]
        for node_name in broken_nodes:
            self.new_addrs[node_name] = ""
        assert_report_item_list_equal(
            config_validators.add_link(
                self.new_addrs,
                {},
                self.coro_nodes,
                self.pcmk_nodes,
                self.existing_link_list,
                self.transport,
                constants.IP_VERSION_64,
            ),
            [
                fixture.error(
                    report_codes.NODE_ADDRESSES_CANNOT_BE_EMPTY,
                    node_name_list=broken_nodes,
                ),
            ],
        )

    def test_used_addrs(self):
        pcmk_nodes = [PacemakerNode("node-remote", "addr-remote")]
        already_existing_addrs = [
            pcmk_nodes[0].addr,
            self.coro_nodes[0].addrs_plain()[0],
        ]
        (
            self.new_addrs["node2"],
            self.new_addrs["node3"],
        ) = already_existing_addrs
        assert_report_item_list_equal(
            config_validators.add_link(
                self.new_addrs,
                {},
                self.coro_nodes,
                pcmk_nodes,
                self.existing_link_list,
                self.transport,
                constants.IP_VERSION_64,
            ),
            [
                fixture.error(
                    report_codes.NODE_ADDRESSES_ALREADY_EXIST,
                    address_list=already_existing_addrs,
                ),
            ],
        )

    def test_unknown_node(self):
        unknown_nodes = ["node4", "node5"]
        for node_name in unknown_nodes:
            self.new_addrs[node_name] = f"{node_name}-addr"
        assert_report_item_list_equal(
            config_validators.add_link(
                self.new_addrs,
                {},
                self.coro_nodes,
                self.pcmk_nodes,
                self.existing_link_list,
                self.transport,
                constants.IP_VERSION_64,
            ),
            [
                fixture.error(
                    report_codes.NODE_NOT_FOUND,
                    node=node_name,
                    searched_types=[],
                )
                for node_name in unknown_nodes
            ],
        )

    def test_duplicate_addrs(self):
        self.new_addrs["node1"] = self.new_addrs["node2"]
        assert_report_item_list_equal(
            config_validators.add_link(
                self.new_addrs,
                {},
                self.coro_nodes,
                self.pcmk_nodes,
                self.existing_link_list,
                self.transport,
                constants.IP_VERSION_64,
            ),
            [
                fixture.error(
                    report_codes.NODE_ADDRESSES_DUPLICATION,
                    address_list=[self.new_addrs["node1"]],
                )
            ],
        )

    def test_unresolvable_addrs(self):
        patch_getaddrinfo(self, list(self.new_addrs.values())[2:])
        assert_report_item_list_equal(
            config_validators.add_link(
                self.new_addrs,
                {},
                self.coro_nodes,
                self.pcmk_nodes,
                self.existing_link_list,
                self.transport,
                constants.IP_VERSION_64,
            ),
            [
                fixture.error(
                    report_codes.NODE_ADDRESSES_UNRESOLVABLE,
                    force_code=report_codes.FORCE,
                    address_list=list(self.new_addrs.values())[:2],
                )
            ],
        )

    def test_unresolvable_addrs_forced(self):
        patch_getaddrinfo(self, list(self.new_addrs.values())[2:])
        assert_report_item_list_equal(
            config_validators.add_link(
                self.new_addrs,
                {},
                self.coro_nodes,
                self.pcmk_nodes,
                self.existing_link_list,
                self.transport,
                constants.IP_VERSION_64,
                force_unresolvable=True,
            ),
            [
                fixture.warn(
                    report_codes.NODE_ADDRESSES_UNRESOLVABLE,
                    address_list=list(self.new_addrs.values())[:2],
                )
            ],
        )

    def test_mixing_ip_families(self):
        new_addrs = {
            "node1": "addr1-new",
            "node2": "10.0.2.2",
            "node3": "::ffff:10:0:2:3",
        }
        assert_report_item_list_equal(
            config_validators.add_link(
                new_addrs,
                {},
                self.coro_nodes,
                self.pcmk_nodes,
                self.existing_link_list,
                self.transport,
                constants.IP_VERSION_64,
            ),
            [
                fixture.error(
                    report_codes.COROSYNC_IP_VERSION_MISMATCH_IN_LINKS,
                    link_numbers=[],
                ),
            ],
        )

    def test_wrong_ip_family_4(self):
        self.new_addrs["node2"] = "::ffff:10:0:2:2"
        assert_report_item_list_equal(
            config_validators.add_link(
                self.new_addrs,
                {},
                self.coro_nodes,
                self.pcmk_nodes,
                self.existing_link_list,
                self.transport,
                constants.IP_VERSION_4,
            ),
            [
                fixture.error(
                    report_codes.COROSYNC_ADDRESS_IP_VERSION_WRONG_FOR_LINK,
                    address=self.new_addrs["node2"],
                    expected_address_type=CorosyncNodeAddressType.IPV4.value,
                    link_number=None,
                ),
            ],
        )

    def test_wrong_ip_family_6(self):
        self.new_addrs["node2"] = "10.0.2.2"
        assert_report_item_list_equal(
            config_validators.add_link(
                self.new_addrs,
                {},
                self.coro_nodes,
                self.pcmk_nodes,
                self.existing_link_list,
                self.transport,
                constants.IP_VERSION_6,
            ),
            [
                fixture.error(
                    report_codes.COROSYNC_ADDRESS_IP_VERSION_WRONG_FOR_LINK,
                    address=self.new_addrs["node2"],
                    expected_address_type=CorosyncNodeAddressType.IPV6.value,
                    link_number=None,
                ),
            ],
        )

    def test_link_options(self):
        assert_report_item_list_equal(
            config_validators.add_link(
                self.new_addrs,
                {
                    "linknumber": "ln",
                    "link_priority": "lp",
                    "mcastport": "mp",
                    "ping_interval": "pi",
                    "ping_precision": "pp",
                    "wrong": "option",
                    "ping_timeout": "pt",
                    "pong_count": "pc",
                    "bad": "option",
                    "transport": "t",
                },
                self.coro_nodes,
                self.pcmk_nodes,
                self.existing_link_list,
                self.transport,
                constants.IP_VERSION_64,
            ),
            [
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_name="linknumber",
                    option_value="ln",
                    allowed_values="0..7",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_name="link_priority",
                    option_value="lp",
                    allowed_values="0..255",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_name="mcastport",
                    option_value="mp",
                    allowed_values="a port number (1..65535)",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_name="ping_interval",
                    option_value="pi",
                    allowed_values="a non-negative integer",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_name="ping_precision",
                    option_value="pp",
                    allowed_values="a non-negative integer",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_name="ping_timeout",
                    option_value="pt",
                    allowed_values="a non-negative integer",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_name="pong_count",
                    option_value="pc",
                    allowed_values="a non-negative integer",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_name="transport",
                    option_value="t",
                    allowed_values=("sctp", "udp"),
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                fixture.error(
                    report_codes.INVALID_OPTIONS,
                    option_names=["bad", "wrong"],
                    option_type="link",
                    allowed=[
                        "link_priority",
                        "linknumber",
                        "mcastport",
                        "ping_interval",
                        "ping_precision",
                        "ping_timeout",
                        "pong_count",
                        "transport",
                    ],
                    allowed_patterns=[],
                ),
            ],
        )

    def test_forbidden_characters(self):
        assert_report_item_list_equal(
            config_validators.add_link(
                {
                    "node1": "addr1-new\n",
                    "node2": "addr2{-}new",
                    "node3": "addr3-new",
                },
                {
                    "linknumber": "\n2",
                    "link_priority": "2\r",
                    "mcastport": "}5405",
                    "ping_interval": "100{",
                    "ping_precision": "\r10\n",
                    "ping_timeout": "{50}",
                    "pong_count": "5\n",
                    "transport": "udp}",
                    "op:.tion": "va}l{ue",
                },
                self.coro_nodes,
                self.pcmk_nodes,
                self.existing_link_list,
                self.transport,
                constants.IP_VERSION_64,
            ),
            [
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_value="addr1-new\n",
                    option_name="node address",
                    **forbidden_characters_kwargs,
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_value="addr2{-}new",
                    option_name="node address",
                    **forbidden_characters_kwargs,
                ),
                fixture.error(
                    report_codes.NODE_ADDRESSES_UNRESOLVABLE,
                    force_code=report_codes.FORCE,
                    address_list=["addr1-new\n", "addr2{-}new"],
                ),
                fixture.error(
                    report_codes.INVALID_OPTIONS,
                    option_names=["op:.tion"],
                    option_type="link",
                    allowed=[
                        "link_priority",
                        "linknumber",
                        "mcastport",
                        "ping_interval",
                        "ping_precision",
                        "ping_timeout",
                        "pong_count",
                        "transport",
                    ],
                    allowed_patterns=[],
                ),
                fixture.error(
                    report_codes.INVALID_USERDEFINED_OPTIONS,
                    option_names=["op:.tion"],
                    option_type="link",
                    allowed_characters="a-z A-Z 0-9 /_-",
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_value="\n2",
                    option_name="linknumber",
                    **forbidden_characters_kwargs,
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_value="2\r",
                    option_name="link_priority",
                    **forbidden_characters_kwargs,
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_value="}5405",
                    option_name="mcastport",
                    **forbidden_characters_kwargs,
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_value="100{",
                    option_name="ping_interval",
                    **forbidden_characters_kwargs,
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_value="\r10\n",
                    option_name="ping_precision",
                    **forbidden_characters_kwargs,
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_value="{50}",
                    option_name="ping_timeout",
                    **forbidden_characters_kwargs,
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_value="5\n",
                    option_name="pong_count",
                    **forbidden_characters_kwargs,
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_value="udp}",
                    option_name="transport",
                    **forbidden_characters_kwargs,
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_value="va}l{ue",
                    option_name="op:.tion",
                    **forbidden_characters_kwargs,
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_value="2\r",
                    option_name="link_priority",
                    allowed_values="0..255",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_value="}5405",
                    option_name="mcastport",
                    allowed_values="a port number (1..65535)",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_value="100{",
                    option_name="ping_interval",
                    allowed_values="a non-negative integer",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_value="\r10\n",
                    option_name="ping_precision",
                    allowed_values="a non-negative integer",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_value="{50}",
                    option_name="ping_timeout",
                    allowed_values="a non-negative integer",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_value="5\n",
                    option_name="pong_count",
                    allowed_values="a non-negative integer",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_value="udp}",
                    option_name="transport",
                    allowed_values=("sctp", "udp"),
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_value="\n2",
                    option_name="linknumber",
                    allowed_values="0..7",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
            ],
        )


class RemoveLinks(TestCase):
    def setUp(self):
        self.existing = ["0", "3", "10", "1", "11"]

    def test_no_link_specified(self):
        assert_report_item_list_equal(
            config_validators.remove_links([], self.existing, "knet"),
            [
                fixture.error(
                    # pylint: disable=line-too-long
                    report_codes.COROSYNC_CANNOT_ADD_REMOVE_LINKS_NO_LINKS_SPECIFIED,
                    add_or_not_remove=False,
                )
            ],
        )

    def _assert_bad_transport(self, transport):
        assert_report_item_list_equal(
            config_validators.remove_links(["3"], self.existing, transport),
            [
                fixture.error(
                    report_codes.COROSYNC_CANNOT_ADD_REMOVE_LINKS_BAD_TRANSPORT,
                    add_or_not_remove=False,
                    actual_transport=transport,
                    required_transports=["knet"],
                )
            ],
        )

    def test_transport_udp(self):
        self._assert_bad_transport("udp")

    def test_transport_udpu(self):
        self._assert_bad_transport("udpu")

    def test_nonexistent_links(self):
        to_remove = ["15", "0", "4", "abc", "1"]
        assert len(to_remove) >= len(self.existing)

        assert_report_item_list_equal(
            config_validators.remove_links(to_remove, self.existing, "knet"),
            [
                fixture.error(
                    report_codes.COROSYNC_LINK_DOES_NOT_EXIST_CANNOT_REMOVE,
                    link_list=sorted(["abc", "4", "15"]),
                    existing_link_list=sorted(["0", "1", "3", "10", "11"]),
                )
            ],
        )

    def test_zero_links_left(self):
        assert_report_item_list_equal(
            config_validators.remove_links(
                self.existing, self.existing, "knet"
            ),
            [
                fixture.error(
                    # pylint: disable=line-too-long
                    report_codes.COROSYNC_CANNOT_ADD_REMOVE_LINKS_TOO_MANY_FEW_LINKS,
                    links_change_count=len(self.existing),
                    links_new_count=0,
                    links_limit_count=1,
                    add_or_not_remove=False,
                )
            ],
        )

    def test_remove_more_than_defined(self):
        assert_report_item_list_equal(
            config_validators.remove_links(
                self.existing + ["2"], self.existing, "knet"
            ),
            [
                fixture.error(
                    # pylint: disable=line-too-long
                    report_codes.COROSYNC_CANNOT_ADD_REMOVE_LINKS_TOO_MANY_FEW_LINKS,
                    # We try to remove more links than defined yet only defined
                    # links are counted here - nonexistent links cannot be
                    # defined so they are not included in the count
                    links_change_count=len(self.existing),
                    # the point of the test is to not get negative number here
                    links_new_count=0,
                    links_limit_count=1,
                    add_or_not_remove=False,
                ),
                fixture.error(
                    report_codes.COROSYNC_LINK_DOES_NOT_EXIST_CANNOT_REMOVE,
                    link_list=["2"],
                    existing_link_list=sorted(["0", "1", "3", "10", "11"]),
                ),
            ],
        )

    def test_duplicate_links(self):
        assert_report_item_list_equal(
            config_validators.remove_links(
                ["abc", "abc", "11", "11", "1", "1", "3"], self.existing, "knet"
            ),
            [
                fixture.error(
                    report_codes.COROSYNC_LINK_NUMBER_DUPLICATION,
                    link_number_list=sorted(["abc", "1", "11"]),
                ),
                fixture.error(
                    report_codes.COROSYNC_LINK_DOES_NOT_EXIST_CANNOT_REMOVE,
                    link_list=["abc"],
                    existing_link_list=sorted(["0", "1", "3", "10", "11"]),
                ),
            ],
        )

    def test_success(self):
        assert_report_item_list_equal(
            config_validators.remove_links(
                ["0", "3", "11"], self.existing, "knet"
            ),
            [],
        )


class UpdateLinkCommon(TestCase):
    # Link update validator is complex, this class tests common cases. For more
    # specific tests see other UpdateLink* classes.
    def test_no_addrs_no_options(self):
        # pylint: disable=no-self-use
        assert_report_item_list_equal(
            config_validators.update_link(
                "0",
                {},
                {},
                {},
                [],
                [],
                ["0"],
                constants.TRANSPORTS_UDP[0],
                constants.IP_VERSION_64,
            ),
            [],
        )

    def test_nonexistent_link(self):
        # pylint: disable=no-self-use
        assert_report_item_list_equal(
            config_validators.update_link(
                "1",
                {},
                {},
                {},
                [],
                [],
                ["0"],
                constants.TRANSPORTS_UDP[0],
                constants.IP_VERSION_64,
            ),
            [
                fixture.error(
                    report_codes.COROSYNC_LINK_DOES_NOT_EXIST_CANNOT_UPDATE,
                    link_number="1",
                    existing_link_list=["0"],
                ),
            ],
        )


class UpdateLinkAddressesMixin:
    def test_swap_addresses(self):
        new_addrs = {
            self.coro_nodes[0].name: (
                self.coro_nodes[1].addr_plain_for_link(self.linknumber)
            ),
            self.coro_nodes[1].name: (
                self.coro_nodes[0].addr_plain_for_link(self.linknumber)
            ),
        }
        patch_getaddrinfo(self, list(new_addrs.values()) + self.existing_addrs)

        assert_report_item_list_equal(
            config_validators.update_link(
                self.linknumber,
                new_addrs,
                {},
                {},
                self.coro_nodes,
                [],
                self.existing_link_list,
                self.transport,
                constants.IP_VERSION_64,
            ),
            [],
        )

    def test_duplicate_new_addresses(self):
        new_addrs = {
            self.coro_nodes[0].name: "addr-new1",
            self.coro_nodes[1].name: "addr-new2",
            self.coro_nodes[2].name: "addr-new1",
        }
        patch_getaddrinfo(self, list(new_addrs.values()) + self.existing_addrs)

        assert_report_item_list_equal(
            config_validators.update_link(
                self.linknumber,
                new_addrs,
                {},
                {},
                self.coro_nodes,
                [],
                self.existing_link_list,
                self.transport,
                constants.IP_VERSION_64,
            ),
            [
                fixture.error(
                    report_codes.NODE_ADDRESSES_DUPLICATION,
                    address_list=["addr-new1"],
                ),
            ],
        )

    def test_remove_address(self):
        patch_getaddrinfo(self, self.existing_addrs)
        assert_report_item_list_equal(
            config_validators.update_link(
                self.linknumber,
                {
                    self.coro_nodes[0].name: "",
                },
                {},
                {},
                self.coro_nodes,
                [],
                self.existing_link_list,
                self.transport,
                constants.IP_VERSION_64,
            ),
            [
                fixture.error(
                    report_codes.NODE_ADDRESSES_CANNOT_BE_EMPTY,
                    node_name_list=[self.coro_nodes[0].name],
                ),
            ],
        )

    def test_address_vs_ip_version(self):
        report_4 = fixture.error(
            report_codes.COROSYNC_ADDRESS_IP_VERSION_WRONG_FOR_LINK,
            address="10.0.0.1",
            expected_address_type=CorosyncNodeAddressType.IPV6.value,
            link_number=self.linknumber,
        )
        report_6 = fixture.error(
            report_codes.COROSYNC_ADDRESS_IP_VERSION_WRONG_FOR_LINK,
            address="::ffff:10:0:2:3",
            expected_address_type=CorosyncNodeAddressType.IPV4.value,
            link_number=self.linknumber,
        )
        test_matrix = (
            (constants.IP_VERSION_4, "10.0.0.1", []),
            (constants.IP_VERSION_4, "::ffff:10:0:2:3", [report_6]),
            (constants.IP_VERSION_6, "10.0.0.1", [report_4]),
            (constants.IP_VERSION_6, "::ffff:10:0:2:3", []),
            (constants.IP_VERSION_46, "10.0.0.1", []),
            (constants.IP_VERSION_46, "::ffff:10:0:2:3", []),
            (constants.IP_VERSION_64, "10.0.0.1", []),
            (constants.IP_VERSION_64, "::ffff:10:0:2:3", []),
            (constants.IP_VERSION_4, "new-addr", []),
            (constants.IP_VERSION_6, "new-addr", []),
            (constants.IP_VERSION_46, "new-addr", []),
            (constants.IP_VERSION_64, "new-addr", []),
        )
        patch_getaddrinfo(self, self.existing_addrs + ["new-addr"])
        for ip_version, new_ip, reports in test_matrix:
            with self.subTest(ip_version=ip_version, new_ip=new_ip):
                assert_report_item_list_equal(
                    config_validators.update_link(
                        self.linknumber,
                        {
                            self.coro_nodes[0].name: new_ip,
                        },
                        {},
                        {},
                        self.coro_nodes,
                        [],
                        self.existing_link_list,
                        self.transport,
                        ip_version,
                    ),
                    reports,
                )

    def test_mixing_ip_families_new_vs_new(self):
        patch_getaddrinfo(self, self.existing_addrs)
        assert_report_item_list_equal(
            config_validators.update_link(
                self.linknumber,
                {
                    self.coro_nodes[0].name: "10.0.0.1",
                    self.coro_nodes[1].name: "::ffff:10:0:2:3",
                },
                {},
                {},
                self.coro_nodes,
                [],
                self.existing_link_list,
                self.transport,
                constants.IP_VERSION_64,
            ),
            [
                fixture.error(
                    report_codes.COROSYNC_IP_VERSION_MISMATCH_IN_LINKS,
                    link_numbers=[],
                ),
            ],
        )

    def test_unresolvable(self):
        patch_getaddrinfo(self, self.existing_addrs)
        assert_report_item_list_equal(
            config_validators.update_link(
                self.linknumber,
                {
                    self.coro_nodes[0].name: "addr-new",
                },
                {},
                {},
                self.coro_nodes,
                [],
                self.existing_link_list,
                self.transport,
                constants.IP_VERSION_64,
            ),
            [
                fixture.error(
                    report_codes.NODE_ADDRESSES_UNRESOLVABLE,
                    force_code=report_codes.FORCE,
                    address_list=["addr-new"],
                ),
            ],
        )

    def test_unresolvable_forced(self):
        patch_getaddrinfo(self, self.existing_addrs)
        assert_report_item_list_equal(
            config_validators.update_link(
                self.linknumber,
                {
                    self.coro_nodes[0].name: "addr-new",
                },
                {},
                {},
                self.coro_nodes,
                [],
                self.existing_link_list,
                self.transport,
                constants.IP_VERSION_64,
                force_unresolvable=True,
            ),
            [
                fixture.warn(
                    report_codes.NODE_ADDRESSES_UNRESOLVABLE,
                    address_list=["addr-new"],
                ),
            ],
        )

    def test_unknown_nodes(self):
        patch_getaddrinfo(self, self.existing_addrs)
        assert_report_item_list_equal(
            config_validators.update_link(
                self.linknumber,
                {
                    "nodeX": "10.0.0.1",
                },
                {},
                {},
                self.coro_nodes,
                [],
                self.existing_link_list,
                self.transport,
                constants.IP_VERSION_64,
            ),
            [
                fixture.error(
                    report_codes.NODE_NOT_FOUND,
                    node="nodeX",
                    searched_types=[],
                )
            ],
        )

    def test_forbidden_characters(self):
        new_addrs = {
            self.coro_nodes[0].name: "addr-new1\n",
            self.coro_nodes[1].name: "}addr-new2{",
            self.coro_nodes[2].name: "addr-new3",
        }
        patch_getaddrinfo(self, list(new_addrs.values()) + self.existing_addrs)

        assert_report_item_list_equal(
            config_validators.update_link(
                self.linknumber,
                new_addrs,
                {},
                {},
                self.coro_nodes,
                [],
                self.existing_link_list,
                self.transport,
                constants.IP_VERSION_64,
            ),
            [
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_value="addr-new1\n",
                    option_name="node address",
                    **forbidden_characters_kwargs,
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_value="}addr-new2{",
                    option_name="node address",
                    **forbidden_characters_kwargs,
                ),
            ],
        )


class UpdateLinkAddressesUdp(UpdateLinkAddressesMixin, TestCase):
    def setUp(self):
        self.linknumber = "0"
        self.transport = constants.TRANSPORTS_UDP[0]
        self.existing_link_list = ["0"]
        self.coro_nodes = [
            node.CorosyncNode(
                f"node{i}", [node.CorosyncNodeAddress(f"addr{i}", "0")], i
            )
            for i in range(1, 5)
        ]
        self.existing_addrs = []
        for a_node in self.coro_nodes:
            self.existing_addrs.extend(a_node.addrs_plain())

    def test_new_address_already_used(self):
        pcmk_nodes = [PacemakerNode("node-remote", "addr-remote")]
        new_addrs = {
            self.coro_nodes[1]
            .name: self.coro_nodes[0]
            .addr_plain_for_link("0"),
            self.coro_nodes[2].name: pcmk_nodes[0].addr,
            self.coro_nodes[3].name: "new-addr",
        }
        patch_getaddrinfo(self, list(new_addrs.values()) + self.existing_addrs)

        assert_report_item_list_equal(
            config_validators.update_link(
                self.linknumber,
                new_addrs,
                {},
                {},
                self.coro_nodes,
                pcmk_nodes,
                self.existing_link_list,
                self.transport,
                constants.IP_VERSION_64,
            ),
            [
                fixture.error(
                    report_codes.NODE_ADDRESSES_ALREADY_EXIST,
                    address_list=["addr-remote", "addr1"],
                ),
            ],
        )

    def test_mixing_ip_families_new_vs_ipv6(self):
        patch_getaddrinfo(self, self.existing_addrs)
        coro_nodes = [
            node.CorosyncNode(
                f"node{i}",
                [node.CorosyncNodeAddress(f"::ffff:10:0:2:{i}", "0")],
                i,
            )
            for i in range(1, 3)
        ]
        assert_report_item_list_equal(
            config_validators.update_link(
                self.linknumber,
                {
                    self.coro_nodes[0].name: "10.0.1.1",
                },
                {},
                {},
                coro_nodes,
                [],
                self.existing_link_list,
                self.transport,
                constants.IP_VERSION_64,
            ),
            [
                fixture.error(
                    report_codes.COROSYNC_IP_VERSION_MISMATCH_IN_LINKS,
                    link_numbers=[],
                ),
            ],
        )

    def test_mixing_ip_families_new_vs_ipv4(self):
        patch_getaddrinfo(self, self.existing_addrs)
        coro_nodes = [
            node.CorosyncNode(
                f"node{i}", [node.CorosyncNodeAddress(f"10.0.0.{i}", "0")], i
            )
            for i in range(1, 3)
        ]
        assert_report_item_list_equal(
            config_validators.update_link(
                self.linknumber,
                {
                    self.coro_nodes[0].name: "::ffff:10:0:3:1",
                },
                {},
                {},
                coro_nodes,
                [],
                self.existing_link_list,
                self.transport,
                constants.IP_VERSION_64,
            ),
            [
                fixture.error(
                    report_codes.COROSYNC_IP_VERSION_MISMATCH_IN_LINKS,
                    link_numbers=[],
                ),
            ],
        )


class UpdateLinkAddressesKnet(UpdateLinkAddressesMixin, TestCase):
    def setUp(self):
        self.linknumber = "1"
        self.transport = constants.TRANSPORTS_KNET[0]
        self.existing_link_list = ["0", "1", "3"]
        self.coro_nodes = [
            node.CorosyncNode(
                f"node{i}",
                [
                    node.CorosyncNodeAddress(f"addr{i}-{j}", f"{j}")
                    for j in self.existing_link_list
                ],
                i,
            )
            for i in range(1, 5)
        ]
        self.existing_addrs = []
        for a_node in self.coro_nodes:
            self.existing_addrs.extend(a_node.addrs_plain())

    def test_new_address_already_used(self):
        pcmk_nodes = [PacemakerNode("node-remote", "addr-remote")]
        new_addrs = {
            self.coro_nodes[0]
            .name: self.coro_nodes[3]
            .addr_plain_for_link("1"),
            self.coro_nodes[1]
            .name: self.coro_nodes[1]
            .addr_plain_for_link("0"),
            self.coro_nodes[2].name: pcmk_nodes[0].addr,
        }
        patch_getaddrinfo(self, list(new_addrs.values()) + self.existing_addrs)

        assert_report_item_list_equal(
            config_validators.update_link(
                "1",
                new_addrs,
                {},
                {},
                self.coro_nodes,
                pcmk_nodes,
                self.existing_link_list,
                self.transport,
                constants.IP_VERSION_64,
            ),
            [
                fixture.error(
                    report_codes.NODE_ADDRESSES_ALREADY_EXIST,
                    address_list=sorted(new_addrs.values()),
                ),
            ],
        )

    def test_mixing_ip_families_new_vs_ipv6(self):
        patch_getaddrinfo(self, self.existing_addrs)
        coro_nodes = [
            node.CorosyncNode(
                f"node{i}",
                [
                    node.CorosyncNodeAddress(f"addr{i}", "0"),
                    node.CorosyncNodeAddress(f"::ffff:10:0:2:{i}", "1"),
                    node.CorosyncNodeAddress(f"10.0.0.{i}", "3"),
                ],
                i,
            )
            for i in range(1, 3)
        ]
        assert_report_item_list_equal(
            config_validators.update_link(
                self.linknumber,
                {
                    self.coro_nodes[0].name: "10.0.1.1",
                },
                {},
                {},
                coro_nodes,
                [],
                self.existing_link_list,
                self.transport,
                constants.IP_VERSION_64,
            ),
            [
                fixture.error(
                    report_codes.COROSYNC_IP_VERSION_MISMATCH_IN_LINKS,
                    link_numbers=[],
                ),
            ],
        )

    def test_mixing_ip_families_new_vs_ipv4(self):
        patch_getaddrinfo(self, self.existing_addrs)
        coro_nodes = [
            node.CorosyncNode(
                f"node{i}",
                [
                    node.CorosyncNodeAddress(f"addr{i}", "0"),
                    node.CorosyncNodeAddress(f"10.0.0.{i}", "1"),
                    node.CorosyncNodeAddress(f"::ffff:10:0:2:{i}", "3"),
                ],
                i,
            )
            for i in range(1, 3)
        ]
        assert_report_item_list_equal(
            config_validators.update_link(
                self.linknumber,
                {
                    self.coro_nodes[0].name: "::ffff:10:0:3:1",
                },
                {},
                {},
                coro_nodes,
                [],
                self.existing_link_list,
                self.transport,
                constants.IP_VERSION_64,
            ),
            [
                fixture.error(
                    report_codes.COROSYNC_IP_VERSION_MISMATCH_IN_LINKS,
                    link_numbers=[],
                ),
            ],
        )


class UpdateLinkKnet(TestCase):
    def test_individual_options_set(self):
        # pylint: disable=no-self-use
        assert_report_item_list_equal(
            config_validators.update_link(
                "2",
                {},
                {
                    "linknumber": "ln",
                    "link_priority": "lp",
                    "mcastport": "mp",
                    "ping_interval": "pi",
                    "ping_precision": "pp",
                    "wrong": "option",
                    "ping_timeout": "pt",
                    "pong_count": "pc",
                    "transport": "t",
                },
                {},
                [],
                [],
                [
                    "2",
                ],
                constants.TRANSPORTS_KNET[0],
                constants.IP_VERSION_64,
            ),
            [
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_name="link_priority",
                    option_value="lp",
                    allowed_values="0..255",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_name="mcastport",
                    option_value="mp",
                    allowed_values="a port number (1..65535)",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_name="ping_interval",
                    option_value="pi",
                    allowed_values="a non-negative integer",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_name="ping_precision",
                    option_value="pp",
                    allowed_values="a non-negative integer",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_name="ping_timeout",
                    option_value="pt",
                    allowed_values="a non-negative integer",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_name="pong_count",
                    option_value="pc",
                    allowed_values="a non-negative integer",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_name="transport",
                    option_value="t",
                    allowed_values=("sctp", "udp"),
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                fixture.error(
                    report_codes.INVALID_OPTIONS,
                    option_names=["linknumber", "wrong"],
                    option_type="link",
                    allowed=[
                        "link_priority",
                        "mcastport",
                        "ping_interval",
                        "ping_precision",
                        "ping_timeout",
                        "pong_count",
                        "transport",
                    ],
                    allowed_patterns=[],
                ),
            ],
        )

    def test_individual_options_unset(self):
        # pylint: disable=no-self-use
        assert_report_item_list_equal(
            config_validators.update_link(
                "2",
                {},
                {
                    "linknumber": "",
                    "link_priority": "",
                    "mcastport": "",
                    "ping_interval": "",
                    "ping_precision": "",
                    "ping_timeout": "",
                    "pong_count": "",
                    "transport": "",
                    "not_valid": "",
                },
                {},
                [],
                [],
                [
                    "2",
                ],
                constants.TRANSPORTS_KNET[0],
                constants.IP_VERSION_64,
            ),
            [
                fixture.error(
                    report_codes.INVALID_OPTIONS,
                    option_names=["linknumber", "not_valid"],
                    option_type="link",
                    allowed=[
                        "link_priority",
                        "mcastport",
                        "ping_interval",
                        "ping_precision",
                        "ping_timeout",
                        "pong_count",
                        "transport",
                    ],
                    allowed_patterns=[],
                ),
            ],
        )

    @staticmethod
    def _prepare_new_option(options, option_name, set_unset):
        if set_unset is False:
            options[option_name] = ""
        elif set_unset is True:
            options[option_name] = "20"

    def test_ping_interval_ping_timeout_dependencies(self):
        test_matrix = (
            # first tuple: a change to be made
            # - ping_interval, ping_timeout
            # - None: no change, False: unset, True: set
            # second tuple: errors expected?
            # - when both were unset, when both were set
            # - we do not consider broken initial status of only one of
            #   timeout, interval set
            ((None, None), (None, None)),
            ((None, False), (None, "interval")),
            ((None, True), ("timeout", None)),
            ((False, None), (None, "timeout")),
            ((False, False), (None, None)),
            ((False, True), ("timeout", "timeout")),
            ((True, None), ("interval", None)),
            ((True, False), ("interval", "interval")),
            ((True, True), (None, None)),
        )
        initial_both_unset = {
            "linknumber": "2",
        }
        initial_both_set = {
            "linknumber": "2",
            "ping_interval": "10",
            "ping_timeout": "10",
        }
        for update_definition, error_definition in test_matrix:
            for error_index, initial_options in enumerate(
                [initial_both_unset, initial_both_set]
            ):
                error_expected = error_definition[error_index]
                new_options = {}
                self._prepare_new_option(
                    new_options, "ping_interval", update_definition[0]
                )
                self._prepare_new_option(
                    new_options, "ping_timeout", update_definition[1]
                )
                with self.subTest(
                    initial_options=initial_options,
                    new_options=new_options,
                    error_expected=error_expected,
                ):
                    reports = []
                    if error_expected and "interval" in error_expected:
                        reports = [
                            fixture.error(
                                report_codes.PREREQUISITE_OPTION_IS_MISSING,
                                option_name="ping_interval",
                                option_type="link",
                                prerequisite_name="ping_timeout",
                                prerequisite_type="link",
                            ),
                        ]
                    if error_expected and "timeout" in error_expected:
                        reports = [
                            fixture.error(
                                report_codes.PREREQUISITE_OPTION_IS_MISSING,
                                option_name="ping_timeout",
                                option_type="link",
                                prerequisite_name="ping_interval",
                                prerequisite_type="link",
                            ),
                        ]
                    assert_report_item_list_equal(
                        config_validators.update_link(
                            "2",
                            {},
                            new_options,
                            initial_options,
                            [],
                            [],
                            [
                                "2",
                            ],
                            constants.TRANSPORTS_KNET[0],
                            constants.IP_VERSION_64,
                        ),
                        reports,
                    )

    def test_ping_interval_ping_timeout_initially_broken(self):
        # pylint: disable=no-self-use
        assert_report_item_list_equal(
            config_validators.update_link(
                "2",
                {},
                {},
                {
                    "linknumber": "2",
                    "ping_timeout": "10",
                },
                [],
                [],
                [
                    "2",
                ],
                constants.TRANSPORTS_KNET[0],
                constants.IP_VERSION_64,
            ),
            [],
        )

    def test_forbidden_characters(self):
        # pylint: disable=no-self-use
        assert_report_item_list_equal(
            config_validators.update_link(
                "2",
                {},
                {
                    "link_priority": "2\r",
                    "mcastport": "}5405",
                    "ping_interval": "100{",
                    "ping_precision": "\r10\n",
                    "ping_timeout": "{50}",
                    "pong_count": "5\n",
                    "transport": "udp}",
                    "op:.tion": "va}l{ue",
                },
                {},
                [],
                [],
                [
                    "2",
                ],
                constants.TRANSPORTS_KNET[0],
                constants.IP_VERSION_64,
            ),
            [
                fixture.error(
                    report_codes.INVALID_OPTIONS,
                    option_names=["op:.tion"],
                    option_type="link",
                    allowed=[
                        "link_priority",
                        "mcastport",
                        "ping_interval",
                        "ping_precision",
                        "ping_timeout",
                        "pong_count",
                        "transport",
                    ],
                    allowed_patterns=[],
                ),
                fixture.error(
                    report_codes.INVALID_USERDEFINED_OPTIONS,
                    option_names=["op:.tion"],
                    option_type="link",
                    allowed_characters="a-z A-Z 0-9 /_-",
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_value="2\r",
                    option_name="link_priority",
                    **forbidden_characters_kwargs,
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_value="}5405",
                    option_name="mcastport",
                    **forbidden_characters_kwargs,
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_value="100{",
                    option_name="ping_interval",
                    **forbidden_characters_kwargs,
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_value="\r10\n",
                    option_name="ping_precision",
                    **forbidden_characters_kwargs,
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_value="{50}",
                    option_name="ping_timeout",
                    **forbidden_characters_kwargs,
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_value="5\n",
                    option_name="pong_count",
                    **forbidden_characters_kwargs,
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_value="udp}",
                    option_name="transport",
                    **forbidden_characters_kwargs,
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_value="va}l{ue",
                    option_name="op:.tion",
                    **forbidden_characters_kwargs,
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_value="2\r",
                    option_name="link_priority",
                    allowed_values="0..255",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_value="}5405",
                    option_name="mcastport",
                    allowed_values="a port number (1..65535)",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_value="100{",
                    option_name="ping_interval",
                    allowed_values="a non-negative integer",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_value="\r10\n",
                    option_name="ping_precision",
                    allowed_values="a non-negative integer",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_value="{50}",
                    option_name="ping_timeout",
                    allowed_values="a non-negative integer",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_value="5\n",
                    option_name="pong_count",
                    allowed_values="a non-negative integer",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_value="udp}",
                    option_name="transport",
                    allowed_values=("sctp", "udp"),
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
            ],
        )


class UpdateLinkUdp(TestCase):
    broadcast_values = (None, "", "0", "1")
    mcastaddr_values = (None, "", "225.1.2.3")
    mcast_error = fixture.error(
        report_codes.PREREQUISITE_OPTION_MUST_BE_DISABLED,
        option_name="mcastaddr",
        option_type="link",
        prerequisite_name="broadcast",
        prerequisite_type="link",
    )

    @staticmethod
    def _fixture_new_values(broadcast, mcastaddr):
        new_values = {}
        if broadcast is not None:
            new_values["broadcast"] = broadcast
        if mcastaddr is not None:
            new_values["mcastaddr"] = mcastaddr
        return new_values

    def test_individual_options_set(self):
        # pylint: disable=no-self-use
        assert_report_item_list_equal(
            config_validators.update_link(
                "0",
                {},
                {
                    "bindnetaddr": "my-network",
                    "broadcast": "yes",
                    "mcastaddr": "my-group",
                    "mcastport": "0",
                    "ttl": "256",
                    "wrong": "value",
                },
                {},
                [],
                [],
                [
                    "0",
                ],
                constants.TRANSPORTS_UDP[0],
                constants.IP_VERSION_4,
            ),
            [
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_value="my-network",
                    option_name="bindnetaddr",
                    allowed_values="an IP address",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_value="yes",
                    option_name="broadcast",
                    allowed_values=("0", "1"),
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_value="my-group",
                    option_name="mcastaddr",
                    allowed_values="an IP address",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_value="0",
                    option_name="mcastport",
                    allowed_values="a port number (1..65535)",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_value="256",
                    option_name="ttl",
                    allowed_values="0..255",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                fixture.error(
                    report_codes.INVALID_OPTIONS,
                    option_names=["wrong"],
                    option_type="link",
                    allowed=[
                        "bindnetaddr",
                        "broadcast",
                        "mcastaddr",
                        "mcastport",
                        "ttl",
                    ],
                    allowed_patterns=[],
                ),
            ],
        )

    def test_individual_options_unset(self):
        # pylint: disable=no-self-use
        assert_report_item_list_equal(
            config_validators.update_link(
                "0",
                {},
                {
                    "bindnetaddr": "",
                    "broadcast": "",
                    "mcastaddr": "",
                    "mcastport": "",
                    "ttl": "",
                    "wrong": "",
                },
                {},
                [],
                [],
                [
                    "0",
                ],
                constants.TRANSPORTS_UDP[0],
                constants.IP_VERSION_4,
            ),
            [
                fixture.error(
                    report_codes.INVALID_OPTIONS,
                    option_names=["wrong"],
                    option_type="link",
                    allowed=[
                        "bindnetaddr",
                        "broadcast",
                        "mcastaddr",
                        "mcastport",
                        "ttl",
                    ],
                    allowed_patterns=[],
                ),
            ],
        )

    def _assert_broadcast_mcast_dependencies(
        self, initial_options, error_expected_for_input
    ):
        for broadcast in self.broadcast_values:
            for mcastaddr in self.mcastaddr_values:
                new_options = self._fixture_new_values(broadcast, mcastaddr)
                error_expected = new_options in error_expected_for_input
                with self.subTest(
                    initial_options=initial_options,
                    new_options=new_options,
                    error_expected=error_expected,
                ):
                    assert_report_item_list_equal(
                        config_validators.update_link(
                            "0",
                            {},
                            new_options,
                            initial_options,
                            [],
                            [],
                            [
                                "0",
                            ],
                            constants.TRANSPORTS_UDP[0],
                            constants.IP_VERSION_4,
                        ),
                        [self.mcast_error] if error_expected else [],
                    )

    def test_broadcast_mcastaddr_dependencies_both_unset(self):
        initial_options = {}
        error_expected_for_input = (
            {"broadcast": "1", "mcastaddr": "225.1.2.3"},
        )
        self._assert_broadcast_mcast_dependencies(
            initial_options, error_expected_for_input
        )

    def test_broadcast_mcastaddr_dependencies_both_disabled(self):
        initial_options = {"broadcast": "0"}
        error_expected_for_input = (
            {"broadcast": "1", "mcastaddr": "225.1.2.3"},
        )
        self._assert_broadcast_mcast_dependencies(
            initial_options, error_expected_for_input
        )

    def test_broadcast_mcastaddr_dependencies_broadcast_enabled(self):
        initial_options = {"broadcast": "1"}
        error_expected_for_input = (
            {"mcastaddr": "225.1.2.3"},
            {"broadcast": "1", "mcastaddr": "225.1.2.3"},
        )
        self._assert_broadcast_mcast_dependencies(
            initial_options, error_expected_for_input
        )

    def test_broadcast_mcastaddr_dependencies_mcastaddr_enabled(self):
        initial_options = {"mcastaddr": "255.2.4.5"}
        error_expected_for_input = (
            {"broadcast": "1"},
            {"broadcast": "1", "mcastaddr": "225.1.2.3"},
        )
        self._assert_broadcast_mcast_dependencies(
            initial_options, error_expected_for_input
        )

    def test_forbidden_characters(self):
        # pylint: disable=no-self-use
        assert_report_item_list_equal(
            config_validators.update_link(
                "0",
                {},
                {
                    "bindnetaddr": "10.1.2.0\n",
                    "broadcast": "1\r",
                    "mcastaddr": "248.251.1.10{",
                    "mcastport": "5405}",
                    "ttl": "128{}",
                    "op:.tion": "va}l{ue",
                },
                {},
                [],
                [],
                [
                    "0",
                ],
                constants.TRANSPORTS_UDP[0],
                constants.IP_VERSION_4,
            ),
            [
                fixture.error(
                    report_codes.INVALID_OPTIONS,
                    option_names=["op:.tion"],
                    option_type="link",
                    allowed=[
                        "bindnetaddr",
                        "broadcast",
                        "mcastaddr",
                        "mcastport",
                        "ttl",
                    ],
                    allowed_patterns=[],
                ),
                fixture.error(
                    report_codes.INVALID_USERDEFINED_OPTIONS,
                    option_names=["op:.tion"],
                    option_type="link",
                    allowed_characters="a-z A-Z 0-9 /_-",
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_value="10.1.2.0\n",
                    option_name="bindnetaddr",
                    **forbidden_characters_kwargs,
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_value="1\r",
                    option_name="broadcast",
                    **forbidden_characters_kwargs,
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_value="248.251.1.10{",
                    option_name="mcastaddr",
                    **forbidden_characters_kwargs,
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_value="5405}",
                    option_name="mcastport",
                    **forbidden_characters_kwargs,
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_value="128{}",
                    option_name="ttl",
                    **forbidden_characters_kwargs,
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_value="va}l{ue",
                    option_name="op:.tion",
                    **forbidden_characters_kwargs,
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_value="10.1.2.0\n",
                    option_name="bindnetaddr",
                    allowed_values="an IP address",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_value="1\r",
                    option_name="broadcast",
                    allowed_values=("0", "1"),
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_value="248.251.1.10{",
                    option_name="mcastaddr",
                    allowed_values="an IP address",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_value="5405}",
                    option_name="mcastport",
                    allowed_values="a port number (1..65535)",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_value="128{}",
                    option_name="ttl",
                    allowed_values="0..255",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
            ],
        )



import dataclasses
from typing import List

from pcs.lib.file.json import JsonExporter
from pcs.lib.interface.config import ExporterInterface

from .types import TokenEntry


class Exporter(ExporterInterface):
    @staticmethod
    def export(config_structure: List[TokenEntry]) -> bytes:
        return JsonExporter.export(
            [dataclasses.asdict(entry) for entry in config_structure]
        )



from collections import defaultdict
from dataclasses import replace as dc_replace
from typing import (
    Dict,
    Iterable,
    List,
    Optional,
    Set,
)

from pcs.common import reports
from pcs.common.types import StringIterable
from pcs.lib import validate
from pcs.lib.external import CommandRunner

from . import const
from .error import (
    ResourceAgentError,
    resource_agent_error_to_report_item,
)
from .name import name_to_void_metadata
from .ocf_transform import ocf_version_to_ocf_unified
from .pcs_transform import (
    get_additional_trace_parameters,
    ocf_unified_to_pcs,
)
from .types import (
    FakeAgentName,
    ResourceAgentMetadata,
    ResourceAgentName,
    ResourceAgentParameter,
)
from .xml import (
    load_fake_agent_metadata,
    load_metadata,
    parse_metadata,
)


class ResourceAgentFacade:
    """
    Provides metadata of and validators for a resource / stonith agent
    """

    def __init__(
        self,
        metadata: ResourceAgentMetadata,
        additional_parameters: Optional[
            Iterable[ResourceAgentParameter]
        ] = None,
    ) -> None:
        """
        metadata -- parsed OCF metadata in a universal format (not version specific)
        additional_parameters -- resource parameters defined outside an agent
        """
        self._raw_metadata = metadata
        self._additional_parameters = additional_parameters
        self._pcs_metadata_cache: Optional[ResourceAgentMetadata] = None

    @property
    def metadata(self) -> ResourceAgentMetadata:
        """
        Return cleaned agent metadata
        """
        if self._pcs_metadata_cache is None:
            self._pcs_metadata_cache = self._get_metadata()
        return self._pcs_metadata_cache

    def _get_metadata(self) -> ResourceAgentMetadata:
        pcs_metadata = ocf_unified_to_pcs(self._raw_metadata)
        if self._additional_parameters:
            pcs_metadata = dc_replace(
                pcs_metadata,
                parameters=(
                    pcs_metadata.parameters + list(self._additional_parameters)
                ),
            )
        return pcs_metadata

    # Facade provides just a basic validation checking that required parameters
    # are set and all set parameters are known to an agent. Missing checks are:
    # 1. values checks - if a param is an integer, then "abc" is not valid
    # 2. errors should be emitted when a deprecated parameter and a
    #    parameter obsoleting it are set at the same time
    # 3. possibly some other checks
    # All of these have been missing in pcs since ever (ad 1. agents have not
    # provided enough info for us to do such validations, ad 3. there were no
    # deprecated parameters before). The checks should be implemented in agents
    # themselves, so we're not adding them now either.

    def get_validators_allowed_parameters(
        self, force: bool = False
    ) -> List[validate.ValidatorInterface]:
        """
        Return validators checking for specified parameters names

        force -- if True, validators produce a warning instead of an error
        """
        return [
            validate.NamesIn(
                {param.name for param in self.metadata.parameters},
                self._validator_option_type,
                severity=reports.item.get_severity(reports.codes.FORCE, force),
            )
        ]

    def get_validators_deprecated_parameters(
        self,
    ) -> List[validate.ValidatorInterface]:
        """
        Return validators looking for deprecated parameters
        """
        # Setting deprecated parameters always emit a warning - we want to allow
        # using them not to break backward compatibility.
        return [
            validate.DeprecatedOption(
                [param.name],
                param.deprecated_by,
                self._validator_option_type,
                severity=reports.ReportItemSeverity.warning(),
            )
            for param in self.metadata.parameters
            if param.deprecated
        ]

    def get_validators_required_parameters(
        self,
        force: bool = False,
        only_parameters: Optional[StringIterable] = None,
    ) -> List[validate.ValidatorInterface]:
        """
        Return validators checking if required parameters were specified

        force -- if True, validators produce a warning instead of an error
        only_parameters -- if set, only specified parameters are checked
        """
        validators: List[validate.ValidatorInterface] = []
        severity = reports.item.get_severity(reports.codes.FORCE, force)
        only_parameters = only_parameters or set()

        required_not_obsoleting: Set[str] = set()
        all_params_deprecated_by = self._get_all_params_deprecated_by()
        for param in self.metadata.parameters:
            if not param.required or param.deprecated:
                continue
            deprecated_by_param = all_params_deprecated_by[param.name]
            if only_parameters and not (
                {param.name} | deprecated_by_param
            ).intersection(only_parameters):
                continue
            if deprecated_by_param:
                validators.append(
                    validate.IsRequiredSome(
                        {param.name} | deprecated_by_param,
                        self._validator_option_type,
                        deprecated_option_name_list=deprecated_by_param,
                        severity=severity,
                    )
                )
            else:
                required_not_obsoleting.add(param.name)

        if required_not_obsoleting:
            validators.append(
                validate.IsRequiredAll(
                    required_not_obsoleting,
                    self._validator_option_type,
                    severity,
                )
            )

        return validators

    @property
    def _validator_option_type(self) -> str:
        return "stonith" if self.metadata.name.is_stonith else "resource"

    def _get_all_params_deprecated_by(self) -> Dict[str, Set[str]]:
        new_olds_map: Dict[str, Set[str]] = defaultdict(set)
        for param in self.metadata.parameters:
            for new_name in param.deprecated_by:
                new_olds_map[new_name].add(param.name)

        result: Dict[str, Set[str]] = defaultdict(set)
        for param in self.metadata.parameters:
            discovered = new_olds_map[param.name]
            while discovered:
                result[param.name] |= discovered
                new_discovered = set()
                for name in discovered:
                    new_discovered |= new_olds_map[name]
                discovered = new_discovered - result[param.name]
        return result


class ResourceAgentFacadeFactory:
    """
    Creates ResourceAgentFacade instances
    """

    def __init__(
        self, runner: CommandRunner, report_processor: reports.ReportProcessor
    ) -> None:
        self._runner = runner
        self._report_processor = report_processor
        self._fenced_metadata: Optional[ResourceAgentMetadata] = None

    def facade_from_parsed_name(
        self, name: ResourceAgentName
    ) -> ResourceAgentFacade:
        """
        Create ResourceAgentFacade based on specified agent name

        name -- agent name to get a facade for
        """
        return self._facade_from_metadata(
            ocf_version_to_ocf_unified(
                parse_metadata(name, load_metadata(self._runner, name))
            )
        )

    def void_facade_from_parsed_name(
        self, name: ResourceAgentName
    ) -> ResourceAgentFacade:
        """
        Create ResourceAgentFacade for a non-existent agent

        name -- name of a non-existent agent to put into the facade
        """
        return self._facade_from_metadata(name_to_void_metadata(name))

    def facade_from_pacemaker_daemon_name(
        self, daemon_name: FakeAgentName
    ) -> ResourceAgentFacade:
        return ResourceAgentFacade(self._get_fake_agent_metadata(daemon_name))

    def _facade_from_metadata(
        self, metadata: ResourceAgentMetadata
    ) -> ResourceAgentFacade:
        additional_parameters = []
        if metadata.name.is_stonith:
            additional_parameters += self._get_fenced_parameters()
        if metadata.name.standard == "ocf" and metadata.name.provider in (
            "heartbeat",
            "pacemaker",
        ):
            additional_parameters += get_additional_trace_parameters(
                metadata.parameters
            )
        return ResourceAgentFacade(metadata, additional_parameters)

    def _get_fake_agent_metadata(
        self, agent_name: FakeAgentName
    ) -> ResourceAgentMetadata:
        return ocf_version_to_ocf_unified(
            parse_metadata(
                ResourceAgentName(const.FAKE_AGENT_STANDARD, None, agent_name),
                load_fake_agent_metadata(self._runner, agent_name),
            )
        )

    def _get_fenced_parameters(self) -> List[ResourceAgentParameter]:
        if self._fenced_metadata is None:
            agent_name = const.PACEMAKER_FENCED
            try:
                self._fenced_metadata = ocf_unified_to_pcs(
                    self._get_fake_agent_metadata(agent_name)
                )
            except ResourceAgentError as e:
                # If pcs is unable to load fenced metadata, cache an empty
                # metadata in order to prevent further futile attempts to load
                # them.
                # Since we are recovering from the failure, we report it as a
                # warning.
                self._report_processor.report(
                    resource_agent_error_to_report_item(
                        e, severity=reports.ReportItemSeverity.warning()
                    )
                )
                self._fenced_metadata = name_to_void_metadata(
                    ResourceAgentName(
                        const.FAKE_AGENT_STANDARD, None, agent_name
                    )
                )
        return self._fenced_metadata.parameters



import logging
import os
from unittest import mock

from pcs.daemon.app import ui
from pcs.lib.auth.provider import AuthProvider
from pcs.lib.auth.types import AuthUser

from pcs_test.tier0.daemon.app import fixtures_app
from pcs_test.tools.misc import get_tmp_dir

LOGIN_BODY = {"username": fixtures_app.USER, "password": fixtures_app.PASSWORD}
PREFIX = "/ui/"

# Don't write errors to test output.
logging.getLogger("tornado.access").setLevel(logging.CRITICAL)


class AppTest(fixtures_app.AppUiTestMixin):
    def setUp(self):
        self.public_dir = get_tmp_dir("tier0_daemon_app_spa")
        self.spa_dir_path = os.path.join(self.public_dir.name, "ui")
        os.makedirs(self.spa_dir_path)
        self.fallback_path = os.path.join(self.public_dir.name, "fallback.html")
        self.index_path = os.path.join(self.spa_dir_path, "index.html")
        self.index_content = "<html/>"
        with open(self.index_path, "w") as index:
            index.write(self.index_content)
        super().setUp()

    def tearDown(self):
        self.public_dir.cleanup()
        super().tearDown()

    def get_routes(self):
        return ui.get_routes(
            url_prefix=PREFIX,
            app_dir=self.spa_dir_path,
            fallback_page_path=self.fallback_path,
            session_storage=self.session_storage,
            auth_provider=AuthProvider(logging.Logger("test logger")),
        )


class Static(AppTest):
    def test_index(self):
        self.assert_success_response(
            self.get(PREFIX),
            self.index_content,
        )


class Fallback(AppTest):
    def setUp(self):
        super().setUp()
        os.remove(self.index_path)
        self.fallback_content = "fallback"
        with open(self.fallback_path, "w") as index:
            index.write(self.fallback_content)

    def test_index(self):
        self.assert_success_response(
            self.get(PREFIX),
            self.fallback_content,
        )


class Login(AppTest):
    def setUp(self):
        super().setUp()
        auth_provider_patcher = mock.patch.object(
            AuthProvider, "auth_by_username_password"
        )
        self.addCleanup(auth_provider_patcher.stop)
        self.auth_provider_mock = auth_provider_patcher.start()

    def test_login_attempt_failed(self):
        self.auth_provider_mock.return_value = None
        self.assert_unauth_ajax(
            self.post(f"{PREFIX}login", LOGIN_BODY, is_ajax=True)
        )
        self.auth_provider_mock.assert_called_once_with(
            fixtures_app.USER, fixtures_app.PASSWORD
        )

    def test_login_attempt_succeeded(self):
        self.auth_provider_mock.return_value = AuthUser(
            fixtures_app.USER, fixtures_app.GROUPS
        )
        response = self.post(f"{PREFIX}login", LOGIN_BODY, is_ajax=True)
        self.assert_success_response(response, "")
        self.auth_provider_mock.assert_called_once_with(
            fixtures_app.USER, fixtures_app.PASSWORD
        )


class Logout(AppTest):
    def test_can_logout(self):
        session1 = self.create_login_session()
        response = self.get(f"{PREFIX}logout", sid=session1.sid, is_ajax=True)
        self.assert_success_response(response, "OK")
        self.assertIsNone(self.session_storage.get(session1.sid))



from unittest import (
    TestCase,
    mock,
)

from pcs.common.reports import codes as report_codes
from pcs.lib.commands import cluster

from pcs_test.tools import fixture
from pcs_test.tools.command_env import get_env_tools

from .common import (
    CLUSTER_UUID,
    fixture_totem,
)


@mock.patch("pcs.lib.commands.cluster.generate_uuid", lambda: CLUSTER_UUID)
class GenerateUuid(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)

    def test_uuid_not_present(self):
        self.config.corosync_conf.load_content(fixture_totem(cluster_uuid=None))
        self.config.env.push_corosync_conf(corosync_conf_text=fixture_totem())
        cluster.generate_cluster_uuid(self.env_assist.get_env())
        self.env_assist.assert_reports([])

    def test_uuid_present(self):
        self.config.corosync_conf.load_content(fixture_totem())
        self.env_assist.assert_raise_library_error(
            lambda: cluster.generate_cluster_uuid(self.env_assist.get_env())
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    report_codes.CLUSTER_UUID_ALREADY_SET,
                    force_code=report_codes.FORCE,
                )
            ]
        )

    def test_uuid_present_with_force(self):
        self.config.corosync_conf.load_content(
            fixture_totem(cluster_uuid="uuid")
        )
        self.config.env.push_corosync_conf(corosync_conf_text=fixture_totem())
        cluster.generate_cluster_uuid(
            self.env_assist.get_env(), force_flags=report_codes.FORCE
        )
        self.env_assist.assert_reports(
            [
                fixture.warn(
                    report_codes.CLUSTER_UUID_ALREADY_SET,
                )
            ]
        )


@mock.patch("pcs.lib.commands.cluster.generate_uuid", lambda: CLUSTER_UUID)
class GenerateUuidLocal(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)

    def test_uuid_not_present(self):
        self.assertEqual(
            cluster.generate_cluster_uuid_local(
                self.env_assist.get_env(),
                fixture_totem(cluster_uuid=None).encode(),
            ),
            fixture_totem().encode(),
        )

    def test_uuid_present(self):
        self.env_assist.assert_raise_library_error(
            lambda: cluster.generate_cluster_uuid_local(
                self.env_assist.get_env(),
                fixture_totem().encode(),
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    report_codes.CLUSTER_UUID_ALREADY_SET,
                    force_code=report_codes.FORCE,
                )
            ]
        )

    def test_uuid_present_with_force(self):
        self.assertEqual(
            cluster.generate_cluster_uuid_local(
                self.env_assist.get_env(),
                fixture_totem(cluster_uuid="uuid").encode(),
                force_flags=report_codes.FORCE,
            ),
            fixture_totem().encode(),
        )
        self.env_assist.assert_reports(
            [
                fixture.warn(
                    report_codes.CLUSTER_UUID_ALREADY_SET,
                )
            ]
        )



from unittest import (
    TestCase,
    mock,
)

from lxml import etree

from pcs import (
    node,
    utils,
)

from pcs_test.tools.assertions import AssertPcsMixin
from pcs_test.tools.cib import get_assert_pcs_effect_mixin
from pcs_test.tools.misc import get_test_resource as rc
from pcs_test.tools.misc import (
    get_tmp_file,
    outdent,
    write_file_to_tmpfile,
)
from pcs_test.tools.pcs_runner import (
    PcsRunner,
    pcs,
)

# pylint: disable=invalid-name, line-too-long

empty_cib = rc("cib-empty-withnodes.xml")


class NodeUtilizationSet(
    TestCase,
    get_assert_pcs_effect_mixin(
        lambda cib: etree.tostring(
            # pylint:disable=undefined-variable
            etree.parse(cib).findall(".//nodes")[0]
        )
    ),
):
    def setUp(self):
        self.temp_cib = get_tmp_file("tier1_node_utilization_set")
        write_file_to_tmpfile(empty_cib, self.temp_cib)
        self.pcs_runner = PcsRunner(self.temp_cib.name)

    def tearDown(self):
        self.temp_cib.close()

    @staticmethod
    def fixture_xml_no_utilization():
        # must match empty_cib
        return """
            <nodes>
                <node id="1" uname="rh7-1" />
                <node id="2" uname="rh7-2" />
            </nodes>
        """

    @staticmethod
    def fixture_xml_empty_utilization():
        # must match empty_cib
        return """
            <nodes>
                <node id="1" uname="rh7-1">
                    <utilization id="nodes-1-utilization" />
                </node>
                <node id="2" uname="rh7-2" />
            </nodes>
        """

    @staticmethod
    def fixture_xml_with_utilization():
        # must match empty_cib
        return """
            <nodes>
                <node id="1" uname="rh7-1">
                    <utilization id="nodes-1-utilization">
                        <nvpair id="nodes-1-utilization-test" name="test"
                            value="100"
                        />
                    </utilization>
                </node>
                <node id="2" uname="rh7-2" />
            </nodes>
        """

    def test_node_utilization_set(self):
        stdout, stderr, retval = pcs(
            self.temp_cib.name, "node utilization rh7-1 test1=10".split()
        )
        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(retval, 0)

        stdout, stderr, retval = pcs(
            self.temp_cib.name, "node utilization rh7-2".split()
        )
        self.assertEqual(
            stdout,
            outdent(
                """\
                Node Utilization:
                """
            ),
        )
        self.assertEqual(stderr, "")
        self.assertEqual(retval, 0)

        stdout, stderr, retval = pcs(
            self.temp_cib.name, "node utilization rh7-1".split()
        )
        self.assertEqual(
            stdout,
            outdent(
                """\
                Node Utilization:
                 rh7-1: test1=10
                """
            ),
        )
        self.assertEqual(stderr, "")
        self.assertEqual(retval, 0)

        stdout, stderr, retval = pcs(
            self.temp_cib.name,
            "node utilization rh7-1 test1=-10 test4=1234".split(),
        )
        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(retval, 0)

        stdout, stderr, retval = pcs(
            self.temp_cib.name, "node utilization rh7-1".split()
        )
        self.assertEqual(
            stdout,
            outdent(
                """\
                Node Utilization:
                 rh7-1: test1=-10 test4=1234
                """
            ),
        )
        self.assertEqual(stderr, "")
        self.assertEqual(retval, 0)

        stdout, stderr, retval = pcs(
            self.temp_cib.name,
            "node utilization rh7-2 test2=321 empty=".split(),
        )
        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(retval, 0)

        stdout, stderr, retval = pcs(
            self.temp_cib.name, "node utilization rh7-2".split()
        )
        self.assertEqual(
            stdout,
            outdent(
                """\
                Node Utilization:
                 rh7-2: test2=321
                """
            ),
        )
        self.assertEqual(stderr, "")
        self.assertEqual(retval, 0)

        stdout, stderr, retval = pcs(
            self.temp_cib.name, "node utilization".split()
        )
        self.assertEqual(
            stdout,
            outdent(
                """\
                Node Utilization:
                 rh7-1: test1=-10 test4=1234
                 rh7-2: test2=321
                """
            ),
        )
        self.assertEqual(stderr, "")
        self.assertEqual(retval, 0)

        stdout, stderr, retval = pcs(
            self.temp_cib.name, "node utilization rh7-2 test1=-20".split()
        )
        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(retval, 0)

        stdout, stderr, retval = pcs(
            self.temp_cib.name, "node utilization --name test1".split()
        )
        self.assertEqual(
            stdout,
            outdent(
                """\
                Node Utilization:
                 rh7-1: test1=-10
                 rh7-2: test1=-20
                """
            ),
        )
        self.assertEqual(stderr, "")
        self.assertEqual(retval, 0)

        stdout, stderr, retval = pcs(
            self.temp_cib.name,
            "node utilization --name test1 rh7-2".split(),
        )
        self.assertEqual(
            stdout,
            outdent(
                """\
                Node Utilization:
                 rh7-2: test1=-20
                """
            ),
        )
        self.assertEqual(stderr, "")
        self.assertEqual(retval, 0)

    def test_refuse_non_option_attribute_parameter_among_options(self):
        self.assert_pcs_fail(
            "node utilization rh7-1 net".split(),
            "Error: missing value of 'net' option\n",
        )

    def test_refuse_option_without_key(self):
        self.assert_pcs_fail(
            "node utilization rh7-1 =1".split(),
            "Error: missing key in '=1' option\n",
        )

    def test_refuse_unknown_node(self):
        self.assert_pcs_fail(
            "node utilization rh7-0 test=10".split(),
            "Error: Unable to find a node: rh7-0\n",
        )

    def test_refuse_value_not_int(self):
        self.assert_pcs_fail(
            "node utilization rh7-1 test1=10 test=int".split(),
            "Error: Value of utilization attribute must be integer: "
            "'test=int'\n",
        )

    def test_keep_empty_nvset(self):
        self.assert_effect(
            "node utilization rh7-1 test=100".split(),
            self.fixture_xml_with_utilization(),
        )
        self.assert_effect(
            "node utilization rh7-1 test=".split(),
            self.fixture_xml_empty_utilization(),
        )

    def test_dont_create_nvset_on_removal(self):
        self.assert_effect(
            "node utilization rh7-1 test=".split(),
            self.fixture_xml_no_utilization(),
        )


class NodeUtilizationPrint(TestCase, AssertPcsMixin):
    def setUp(self):
        self.temp_cib = get_tmp_file("tier1_node_utilization_print")
        write_file_to_tmpfile(empty_cib, self.temp_cib)
        self.pcs_runner = PcsRunner(self.temp_cib.name)

    def tearDown(self):
        self.temp_cib.close()

    @mock.patch("pcs.node.utils")
    def test_refuse_when_node_not_in_cib_and_is_not_remote(self, mock_utils):
        mock_cib = mock.MagicMock()
        mock_cib.getElementsByTagName = mock.Mock(return_value=[])

        mock_utils.get_cib_dom = mock.Mock(return_value=mock_cib)
        mock_utils.usefile = False
        mock_utils.getNodeAttributesFromPacemaker = mock.Mock(return_value=[])
        mock_utils.err = mock.Mock(side_effect=SystemExit)

        self.assertRaises(
            SystemExit, lambda: node.print_node_utilization("some")
        )

    def test_refuse_when_node_not_in_mocked_cib(self):
        self.assert_pcs_fail(
            "node utilization some_nonexistent_node".split(),
            [
                "Error: Unable to find a node: some_nonexistent_node",
            ],
        )


class NodeStandby(TestCase, AssertPcsMixin):
    def setUp(self):
        self.temp_cib = get_tmp_file("tier1_node_standby")
        write_file_to_tmpfile(rc("cib-empty-with3nodes.xml"), self.temp_cib)
        self.pcs_runner = PcsRunner(self.temp_cib.name)

    def tearDown(self):
        self.temp_cib.close()

    def fixture_standby_all(self):
        self.assert_pcs_success("node standby --all".split())
        self.assert_standby_all()

    def assert_standby_none(self):
        self.assert_pcs_success("node attribute".split(), "Node Attributes:\n")

    def assert_standby_all(self):
        self.assert_pcs_success(
            "node attribute".split(),
            outdent(
                """\
                Node Attributes:
                 rh7-1: standby=on
                 rh7-2: standby=on
                 rh7-3: standby=on
                """
            ),
        )

    def test_local_node(self):
        self.assert_standby_none()
        self.assert_pcs_fail(
            "node standby".split(),
            "Error: Node(s) must be specified if -f is used\n",
        )
        self.assert_standby_none()

        self.fixture_standby_all()
        self.assert_pcs_fail(
            "node unstandby".split(),
            "Error: Node(s) must be specified if -f is used\n",
        )
        self.assert_standby_all()

    def test_one_bad_node(self):
        self.assert_standby_none()
        self.assert_pcs_fail(
            "node standby nonexistent-node".split(),
            "Error: Node 'nonexistent-node' does not appear to exist in configuration\n",
        )
        self.assert_standby_none()

        self.fixture_standby_all()
        self.assert_pcs_fail(
            "node unstandby nonexistent-node".split(),
            "Error: Node 'nonexistent-node' does not appear to exist in configuration\n",
        )
        self.assert_standby_all()

    def test_bad_node_cancels_all_changes(self):
        self.assert_standby_none()
        self.assert_pcs_fail(
            "node standby rh7-1 nonexistent-node and-another rh7-2".split(),
            "Error: Node 'nonexistent-node' does not appear to exist in configuration\n"
            "Error: Node 'and-another' does not appear to exist in configuration\n",
        )
        self.assert_standby_none()

        self.fixture_standby_all()
        self.assert_pcs_fail(
            "node standby rh7-1 nonexistent-node and-another rh7-2".split(),
            "Error: Node 'nonexistent-node' does not appear to exist in configuration\n"
            "Error: Node 'and-another' does not appear to exist in configuration\n",
        )
        self.assert_standby_all()

    def test_all_nodes(self):
        self.assert_standby_none()
        self.assert_pcs_success("node standby --all".split())
        self.fixture_standby_all()

        self.assert_pcs_success("node unstandby --all".split())
        self.assert_standby_none()

    def test_one_node_with_repeat(self):
        self.assert_standby_none()
        self.assert_pcs_success("node standby rh7-1".split())
        self.assert_pcs_success(
            "node attribute".split(),
            outdent(
                """\
                Node Attributes:
                 rh7-1: standby=on
                """
            ),
        )
        self.assert_pcs_success("node standby rh7-1".split())

        self.fixture_standby_all()
        self.assert_pcs_success("node unstandby rh7-1".split())
        self.assert_pcs_success(
            "node attribute".split(),
            outdent(
                """\
                Node Attributes:
                 rh7-2: standby=on
                 rh7-3: standby=on
                """
            ),
        )
        self.assert_pcs_success("node unstandby rh7-1".split())

    def test_more_nodes(self):
        self.assert_standby_none()
        self.assert_pcs_success("node standby rh7-1 rh7-2".split())
        self.assert_pcs_success(
            "node attribute".split(),
            outdent(
                """\
                Node Attributes:
                 rh7-1: standby=on
                 rh7-2: standby=on
                """
            ),
        )

        self.fixture_standby_all()
        self.assert_pcs_success("node unstandby rh7-1 rh7-2".split())
        self.assert_pcs_success(
            "node attribute".split(),
            outdent(
                """\
                Node Attributes:
                 rh7-3: standby=on
                """
            ),
        )

    def test_all_and_nodelist(self):
        self.assert_pcs_fail(
            "node standby rh7-1 rh7-2 --all".split(),
            "Error: Cannot specify both --all and a list of nodes.\n",
        )
        self.assert_pcs_fail(
            "node unstandby rh7-1 rh7-2 --all".split(),
            "Error: Cannot specify both --all and a list of nodes.\n",
        )


class NodeMaintenance(TestCase, AssertPcsMixin):
    def setUp(self):
        self.temp_cib = get_tmp_file("tier1_node_maintenance")
        write_file_to_tmpfile(rc("cib-empty-with3nodes.xml"), self.temp_cib)
        self.pcs_runner = PcsRunner(self.temp_cib.name)

    def tearDown(self):
        self.temp_cib.close()

    def fixture_maintenance_all(self):
        self.assert_pcs_success("node maintenance --all".split())
        self.assert_maintenance_all()

    def assert_maintenance_none(self):
        self.assert_pcs_success("node attribute".split(), "Node Attributes:\n")

    def assert_maintenance_all(self):
        self.assert_pcs_success(
            "node attribute".split(),
            outdent(
                """\
                Node Attributes:
                 rh7-1: maintenance=on
                 rh7-2: maintenance=on
                 rh7-3: maintenance=on
                """
            ),
        )

    def test_local_node(self):
        self.assert_maintenance_none()
        self.assert_pcs_fail(
            "node maintenance".split(),
            "Error: Node(s) must be specified if -f is used\n",
        )
        self.assert_maintenance_none()

        self.fixture_maintenance_all()
        self.assert_pcs_fail(
            "node unmaintenance".split(),
            "Error: Node(s) must be specified if -f is used\n",
        )
        self.assert_maintenance_all()

    def test_one_bad_node(self):
        self.assert_maintenance_none()
        self.assert_pcs_fail(
            "node maintenance nonexistent-node".split(),
            "Error: Node 'nonexistent-node' does not appear to exist in configuration\n",
        )
        self.assert_maintenance_none()

        self.fixture_maintenance_all()
        self.assert_pcs_fail(
            "node unmaintenance nonexistent-node".split(),
            "Error: Node 'nonexistent-node' does not appear to exist in configuration\n",
        )
        self.assert_maintenance_all()

    def test_bad_node_cancels_all_changes(self):
        self.assert_maintenance_none()
        self.assert_pcs_fail(
            "node maintenance rh7-1 nonexistent-node and-another rh7-2".split(),
            "Error: Node 'nonexistent-node' does not appear to exist in configuration\n"
            "Error: Node 'and-another' does not appear to exist in configuration\n",
        )
        self.assert_maintenance_none()

        self.fixture_maintenance_all()
        self.assert_pcs_fail(
            "node maintenance rh7-1 nonexistent-node and-another rh7-2".split(),
            "Error: Node 'nonexistent-node' does not appear to exist in configuration\n"
            "Error: Node 'and-another' does not appear to exist in configuration\n",
        )
        self.assert_maintenance_all()

    def test_all_nodes(self):
        self.assert_maintenance_none()
        self.assert_pcs_success("node maintenance --all".split())
        self.fixture_maintenance_all()

        self.assert_pcs_success("node unmaintenance --all".split())
        self.assert_maintenance_none()

    def test_one_node_with_repeat(self):
        self.assert_maintenance_none()
        self.assert_pcs_success("node maintenance rh7-1".split())
        self.assert_pcs_success(
            "node attribute".split(),
            outdent(
                """\
                Node Attributes:
                 rh7-1: maintenance=on
                """
            ),
        )
        self.assert_pcs_success("node maintenance rh7-1".split())

        self.fixture_maintenance_all()
        self.assert_pcs_success("node unmaintenance rh7-1".split())
        self.assert_pcs_success(
            "node attribute".split(),
            outdent(
                """\
                Node Attributes:
                 rh7-2: maintenance=on
                 rh7-3: maintenance=on
                """
            ),
        )
        self.assert_pcs_success("node unmaintenance rh7-1".split())

    def test_more_nodes(self):
        self.assert_maintenance_none()
        self.assert_pcs_success("node maintenance rh7-1 rh7-2".split())
        self.assert_pcs_success(
            "node attribute".split(),
            outdent(
                """\
                Node Attributes:
                 rh7-1: maintenance=on
                 rh7-2: maintenance=on
                """
            ),
        )

        self.fixture_maintenance_all()
        self.assert_pcs_success("node unmaintenance rh7-1 rh7-2".split())
        self.assert_pcs_success(
            "node attribute".split(),
            outdent(
                """\
                Node Attributes:
                 rh7-3: maintenance=on
                """
            ),
        )

    def test_all_and_nodelist(self):
        self.assert_pcs_fail(
            "node maintenance rh7-1 rh7-2 --all".split(),
            "Error: Cannot specify both --all and a list of nodes.\n",
        )
        self.assert_pcs_fail(
            "node unmaintenance rh7-1 rh7-2 --all".split(),
            "Error: Cannot specify both --all and a list of nodes.\n",
        )


class NodeAttributeTest(
    TestCase,
    get_assert_pcs_effect_mixin(
        lambda cib: etree.tostring(
            # pylint:disable=undefined-variable
            etree.parse(cib).findall(".//nodes")[0]
        )
    ),
):
    # pylint: disable=too-many-public-methods
    def setUp(self):
        self.temp_cib = get_tmp_file("tier1_node_attribute")
        write_file_to_tmpfile(empty_cib, self.temp_cib)
        self.pcs_runner = PcsRunner(self.temp_cib.name)

    def tearDown(self):
        self.temp_cib.close()

    def fixture_attrs(self, nodes, attrs=None):
        attrs = {} if attrs is None else attrs
        xml_lines = ["<nodes>"]
        for node_id, node_name in enumerate(nodes, 1):
            xml_lines.extend(
                [
                    '<node id="{0}" uname="{1}">'.format(node_id, node_name),
                    '<instance_attributes id="nodes-{0}">'.format(node_id),
                ]
            )
            nv = '<nvpair id="nodes-{id}-{name}" name="{name}" value="{val}"/>'
            for name, value in attrs.get(node_name, {}).items():
                xml_lines.append(nv.format(id=node_id, name=name, val=value))
            xml_lines.extend(["</instance_attributes>", "</node>"])
        xml_lines.append("</nodes>")

        utils_usefile_original = utils.usefile
        utils_filename_original = utils.filename
        utils.usefile = True
        utils.filename = self.temp_cib.name
        output, retval = utils.run(
            ["cibadmin", "--modify", "--xml-text", "\n".join(xml_lines)]
        )
        utils.usefile = utils_usefile_original
        utils.filename = utils_filename_original
        assert output == ""
        assert retval == 0

    @staticmethod
    def fixture_xml_no_attrs():
        # must match empty_cib
        return """
            <nodes>
                <node id="1" uname="rh7-1" />
                <node id="2" uname="rh7-2" />
            </nodes>
        """

    @staticmethod
    def fixture_xml_empty_attrs():
        # must match empty_cib
        return """
            <nodes>
                <node id="1" uname="rh7-1">
                    <instance_attributes id="nodes-1" />
                </node>
                <node id="2" uname="rh7-2" />
            </nodes>
        """

    @staticmethod
    def fixture_xml_with_attrs():
        # must match empty_cib
        return """
            <nodes>
                <node id="1" uname="rh7-1">
                    <instance_attributes id="nodes-1">
                        <nvpair id="nodes-1-test" name="test" value="100" />
                    </instance_attributes>
                </node>
                <node id="2" uname="rh7-2" />
            </nodes>
        """

    def test_show_empty(self):
        self.fixture_attrs(["rh7-1", "rh7-2"])
        self.assert_pcs_success("node attribute".split(), "Node Attributes:\n")

    def test_show_nonempty(self):
        self.fixture_attrs(
            ["rh7-1", "rh7-2"],
            {
                "rh7-1": {
                    "IP": "192.168.1.1",
                },
                "rh7-2": {
                    "IP": "192.168.1.2",
                },
            },
        )
        self.assert_pcs_success(
            "node attribute".split(),
            """\
Node Attributes:
 rh7-1: IP=192.168.1.1
 rh7-2: IP=192.168.1.2
""",
        )

    def test_show_multiple_per_node(self):
        self.fixture_attrs(
            ["rh7-1", "rh7-2"],
            {
                "rh7-1": {
                    "IP": "192.168.1.1",
                    "alias": "node1",
                },
                "rh7-2": {
                    "IP": "192.168.1.2",
                    "alias": "node2",
                },
            },
        )
        self.assert_pcs_success(
            "node attribute".split(),
            """\
Node Attributes:
 rh7-1: IP=192.168.1.1 alias=node1
 rh7-2: IP=192.168.1.2 alias=node2
""",
        )

    def test_show_one_node(self):
        self.fixture_attrs(
            ["rh7-1", "rh7-2"],
            {
                "rh7-1": {
                    "IP": "192.168.1.1",
                    "alias": "node1",
                },
                "rh7-2": {
                    "IP": "192.168.1.2",
                    "alias": "node2",
                },
            },
        )
        self.assert_pcs_success(
            "node attribute rh7-1".split(),
            """\
Node Attributes:
 rh7-1: IP=192.168.1.1 alias=node1
""",
        )

    def test_show_missing_node(self):
        self.fixture_attrs(
            ["rh7-1", "rh7-2"],
            {
                "rh7-1": {
                    "IP": "192.168.1.1",
                    "alias": "node1",
                },
                "rh7-2": {
                    "IP": "192.168.1.2",
                    "alias": "node2",
                },
            },
        )
        self.assert_pcs_success(
            "node attribute rh7-3".split(),
            """\
Node Attributes:
""",
        )

    def test_show_name(self):
        self.fixture_attrs(
            ["rh7-1", "rh7-2"],
            {
                "rh7-1": {
                    "IP": "192.168.1.1",
                    "alias": "node1",
                },
                "rh7-2": {
                    "IP": "192.168.1.2",
                    "alias": "node2",
                },
            },
        )
        self.assert_pcs_success(
            "node attribute --name alias".split(),
            """\
Node Attributes:
 rh7-1: alias=node1
 rh7-2: alias=node2
""",
        )

    def test_show_missing_name(self):
        self.fixture_attrs(
            ["rh7-1", "rh7-2"],
            {
                "rh7-1": {
                    "IP": "192.168.1.1",
                    "alias": "node1",
                },
                "rh7-2": {
                    "IP": "192.168.1.2",
                    "alias": "node2",
                },
            },
        )
        self.assert_pcs_success(
            "node attribute --name missing".split(),
            """\
Node Attributes:
""",
        )

    def test_show_node_and_name(self):
        self.fixture_attrs(
            ["rh7-1", "rh7-2"],
            {
                "rh7-1": {
                    "IP": "192.168.1.1",
                    "alias": "node1",
                },
                "rh7-2": {
                    "IP": "192.168.1.2",
                    "alias": "node2",
                },
            },
        )
        self.assert_pcs_success(
            "node attribute --name alias rh7-1".split(),
            """\
Node Attributes:
 rh7-1: alias=node1
""",
        )

    def test_set_new(self):
        self.fixture_attrs(["rh7-1", "rh7-2"])
        self.assert_pcs_success("node attribute rh7-1 IP=192.168.1.1".split())
        self.assert_pcs_success(
            "node attribute".split(),
            """\
Node Attributes:
 rh7-1: IP=192.168.1.1
""",
        )
        self.assert_pcs_success("node attribute rh7-2 IP=192.168.1.2".split())
        self.assert_pcs_success(
            "node attribute".split(),
            """\
Node Attributes:
 rh7-1: IP=192.168.1.1
 rh7-2: IP=192.168.1.2
""",
        )

    def test_set_existing(self):
        self.fixture_attrs(
            ["rh7-1", "rh7-2"],
            {
                "rh7-1": {
                    "IP": "192.168.1.1",
                },
                "rh7-2": {
                    "IP": "192.168.1.2",
                },
            },
        )
        self.assert_pcs_success("node attribute rh7-2 IP=192.168.2.2".split())
        self.assert_pcs_success(
            "node attribute".split(),
            """\
Node Attributes:
 rh7-1: IP=192.168.1.1
 rh7-2: IP=192.168.2.2
""",
        )

    def test_unset(self):
        self.fixture_attrs(
            ["rh7-1", "rh7-2"],
            {
                "rh7-1": {
                    "IP": "192.168.1.1",
                },
                "rh7-2": {
                    "IP": "192.168.1.2",
                },
            },
        )
        self.assert_pcs_success("node attribute rh7-2 IP=".split())
        self.assert_pcs_success(
            "node attribute".split(),
            """\
Node Attributes:
 rh7-1: IP=192.168.1.1
""",
        )

    def test_unset_nonexisting(self):
        self.fixture_attrs(
            ["rh7-1", "rh7-2"],
            {
                "rh7-1": {
                    "IP": "192.168.1.1",
                },
                "rh7-2": {
                    "IP": "192.168.1.2",
                },
            },
        )
        self.assert_pcs_result(
            "node attribute rh7-1 missing=".split(),
            stdout_full="",
            stderr_full="Error: attribute: 'missing' doesn't exist for node: 'rh7-1'\n",
            returncode=2,
        )

    def test_unset_nonexisting_forced(self):
        self.fixture_attrs(
            ["rh7-1", "rh7-2"],
            {
                "rh7-1": {
                    "IP": "192.168.1.1",
                },
                "rh7-2": {
                    "IP": "192.168.1.2",
                },
            },
        )
        self.assert_pcs_success(
            "node attribute rh7-1 missing= --force".split(), ""
        )

    def test_keep_empty_nvset(self):
        self.assert_effect(
            "node attribute rh7-1 test=100".split(),
            self.fixture_xml_with_attrs(),
        )
        self.assert_effect(
            "node attribute rh7-1 test=".split(),
            self.fixture_xml_empty_attrs(),
        )

    def test_dont_create_nvset_on_removal(self):
        # pcs does not actually do cib editing, it passes it to crm_node. So
        # this behaves differently than the rest of pcs - instead of doing
        # nothing it returns an error.
        # Should be changed to be consistent with the rest of pcs.
        stdout, stderr, retval = pcs(
            self.temp_cib.name, "node attribute rh7-1 test=".split()
        )
        self.assertEqual(stdout, "")
        self.assertEqual(
            stderr, "Error: attribute: 'test' doesn't exist for node: 'rh7-1'\n"
        )
        self.assertEqual(retval, 2)



from unittest import (
    TestCase,
    mock,
)

from pcs.lib.auth.config.facade import Facade
from pcs.lib.auth.config.types import TokenEntry


def _token_fixture(token):
    return TokenEntry(
        token=token, username=f"user-{token}", creation_date="now"
    )


class FacadeTest(TestCase):
    def setUp(self):
        self.tokens = tuple(f"token{i}" for i in range(3))
        self.facade = Facade(
            tuple(_token_fixture(token) for token in self.tokens)
        )

    def test_non_existing_token(self):
        self.assertIsNone(self.facade.get_user("token3"))

    def test_valid_token(self):
        self.assertEqual(self.facade.get_user("token1"), "user-token1")

    @mock.patch("pcs.lib.auth.config.facade.get_unique_uuid")
    def test_add_user(self, uuid_mock):
        uuid = "unique uuid"
        username = "new_user"
        uuid_mock.return_value = uuid
        self.assertEqual(uuid, self.facade.add_user(username))
        uuid_mock.assert_called_once_with(self.tokens)
        token_entry = self.facade.config[-1]
        self.assertEqual(uuid, token_entry.token)
        self.assertEqual(username, token_entry.username)

    @mock.patch("pcs.lib.auth.config.facade.get_unique_uuid")
    def test_add_existing_user(self, uuid_mock):
        uuid = "unique uuid"
        username = "user-token2"
        self.assertEqual(self.facade.get_user("token2"), username)
        uuid_mock.return_value = uuid
        self.assertEqual(uuid, self.facade.add_user(username))
        uuid_mock.assert_called_once_with(self.tokens)
        token_entry = self.facade.config[-1]
        self.assertEqual(uuid, token_entry.token)
        self.assertEqual(username, token_entry.username)



from pcs.daemon.app.common import RedirectHandler

from pcs_test.tier0.daemon.app.fixtures_app import AppTest

MANAGE = "/manage"


class RedirectTest(AppTest):
    def get_routes(self):
        return [(r"/", RedirectHandler, dict(url=MANAGE))]

    def test_redirects_to_given_location(self):
        response = self.get("/")
        self.assert_headers_contains(response.headers, {"Location": MANAGE})
        self.assertEqual(response.code, 301)



from dataclasses import (
    dataclass,
    field,
)
from typing import (
    List,
    Optional,
)

from pcs.common.interface.dto import (
    DataTransferObject,
    meta,
)


@dataclass(frozen=True)
class ResourceAgentNameDto(DataTransferObject):
    standard: str
    provider: Optional[str]
    type: str


def get_resource_agent_full_name(agent_name: ResourceAgentNameDto) -> str:
    return ":".join(
        filter(
            None, [agent_name.standard, agent_name.provider, agent_name.type]
        )
    )


@dataclass(frozen=True)
class ListResourceAgentNameDto(DataTransferObject):
    names: List[ResourceAgentNameDto]


@dataclass(frozen=True)
class ResourceAgentActionDto(DataTransferObject):
    # pylint: disable=too-many-instance-attributes

    # (start, stop, promote...), mandatory by both OCF 1.0 and 1.1
    name: str
    # mandatory by both OCF 1.0 and 1.1, sometimes not defined by agents
    timeout: Optional[str]
    # optional by both OCF 1.0 and 1.1
    interval: Optional[str]
    # optional by OCF 1.1
    # not allowed by OCF 1.0, defined in OCF 1.0 agents anyway
    role: Optional[str]
    # OCF name: 'start-delay', optional by both OCF 1.0 and 1.1
    start_delay: Optional[str] = field(metadata=meta(name="start-delay"))
    # optional by both OCF 1.0 and 1.1
    depth: Optional[str]
    # not allowed by any OCF, defined in OCF 1.0 agents anyway
    automatic: bool
    # not allowed by any OCF, defined in OCF 1.0 agents anyway
    on_target: bool


@dataclass(frozen=True)
class ResourceAgentParameterDto(DataTransferObject):
    # pylint: disable=too-many-instance-attributes

    # name of the parameter
    name: str
    # short description
    shortdesc: Optional[str]
    # long description
    longdesc: Optional[str]
    # data type of the parameter
    type: str
    # default value of the parameter
    default: Optional[str]
    # allowed values, only defined if type == 'select'
    enum_values: Optional[List[str]]
    # True if it is a required parameter, False otherwise
    required: bool
    # True if the parameter is meant for advanced users
    advanced: bool
    # True if the parameter is deprecated, False otherwise
    deprecated: bool
    # list of parameters deprecating this one
    deprecated_by: List[str]
    # text describing / explaining the deprecation
    deprecated_desc: Optional[str]
    # should the parameter's value be unique across same agent resources?
    unique_group: Optional[str]
    # changing this parameter's value triggers a reload instead of a restart
    reloadable: bool


@dataclass(frozen=True)
class ResourceAgentMetadataDto(DataTransferObject):
    name: ResourceAgentNameDto
    shortdesc: Optional[str]
    longdesc: Optional[str]
    parameters: List[ResourceAgentParameterDto]
    actions: List[ResourceAgentActionDto]



from typing import (
    Callable,
    Iterable,
    List,
    Optional,
    Set,
)

from pcs.cli.common.tools import print_to_stderr
from pcs.common.reports import (
    ReportItem,
    ReportItemSeverity,
    ReportProcessor,
)
from pcs.common.reports.dto import ReportItemDto
from pcs.common.reports.types import SeverityLevel

from .messages import report_item_msg_from_dto
from .output import (
    add_context_to_message,
    deprecation_warning,
    error,
    warn,
)

ReportItemPreprocessor = Callable[[ReportItem], Optional[ReportItem]]


class ReportProcessorToConsole(ReportProcessor):
    def __init__(self, debug: bool = False) -> None:
        super().__init__()
        self.debug = debug
        self._ignore_severities = self._get_ignored_severities([])
        self._report_item_preprocessor: ReportItemPreprocessor = lambda x: x

    def _do_report(self, report_item: ReportItem) -> None:
        filtered_report_item = self._report_item_preprocessor(report_item)
        if not filtered_report_item:
            return
        report_item_dto = filtered_report_item.to_dto()
        if report_item_dto.severity.level not in self._ignore_severities:
            print_report(report_item_dto)

    def _get_ignored_severities(
        self, suppressed_severity_list: Iterable[SeverityLevel]
    ) -> Set[SeverityLevel]:
        ignore_severities = set(suppressed_severity_list)
        if self.debug:
            ignore_severities -= {ReportItemSeverity.DEBUG}
        else:
            ignore_severities |= {ReportItemSeverity.DEBUG}
        return ignore_severities

    def suppress_reports_of_severity(
        self, severity_list: Iterable[SeverityLevel]
    ) -> None:
        self._ignore_severities = self._get_ignored_severities(severity_list)

    def set_report_item_preprocessor(
        self,
        report_item_preprocessor: ReportItemPreprocessor,
    ) -> None:
        self._report_item_preprocessor = report_item_preprocessor


def print_report(report_item_dto: ReportItemDto) -> None:
    cli_report_msg = report_item_msg_from_dto(report_item_dto.message)
    msg = cli_report_msg.message
    if not msg:
        return
    msg = add_context_to_message(msg, report_item_dto.context)

    severity = report_item_dto.severity.level
    if severity == ReportItemSeverity.ERROR:
        error(
            add_context_to_message(
                cli_report_msg.get_message_with_force_text(
                    report_item_dto.severity.force_code
                ),
                report_item_dto.context,
            )
        )
    elif severity == ReportItemSeverity.WARNING:
        warn(msg)
    elif severity == ReportItemSeverity.DEPRECATION:
        deprecation_warning(msg)
    else:
        print_to_stderr(msg)


def has_errors(report_list_dto: List[ReportItemDto]) -> bool:
    for report_item_dto in report_list_dto:
        if _is_error(report_item_dto):
            return True
    return False


def _is_error(report_item_dto: ReportItemDto) -> bool:
    return report_item_dto.severity.level == ReportItemSeverity.ERROR



from multiprocessing import Queue
from unittest import (
    TestCase,
    mock,
)

from pcs.common import reports
from pcs.common.async_tasks import types
from pcs.common.async_tasks.dto import (
    CommandDto,
    CommandOptionsDto,
)
from pcs.daemon.async_tasks.types import Command
from pcs.daemon.async_tasks.worker import executor
from pcs.daemon.async_tasks.worker.types import (
    Message,
    TaskExecuted,
    TaskFinished,
    WorkerCommand,
)

from .dummy_commands import (
    RESULT,
    test_api_v1_compatibility_mode,
    test_command_map,
)
from .helpers import (
    AUTH_USER,
    MockOsKillMixin,
    PermissionsCheckerMock,
)

TASK_IDENT = "id0"
WORKER_PID = 2222
COMMAND_OPTIONS = CommandOptionsDto(request_timeout=None)


executor.worker_com = Queue()  # patched at runtime


@mock.patch(
    "pcs.daemon.async_tasks.worker.executor.COMMAND_MAP", test_command_map
)
@mock.patch(
    "pcs.daemon.async_tasks.worker.executor.API_V1_COMPATIBILITY_MODE",
    test_api_v1_compatibility_mode,
)
@mock.patch(
    "pcs.daemon.async_tasks.worker.executor.getLogger", mock.MagicMock()
)
@mock.patch(
    "pcs.daemon.async_tasks.worker.executor.PermissionsChecker",
    lambda _: PermissionsCheckerMock({}),
)
@mock.patch("os.getpid")
class TestExecutor(MockOsKillMixin, TestCase):
    """
    Tests the test_executor function

    Every test needs its own queue to avoid mixing messages from other tests
    because tests are running concurrently
    """

    def setUp(self) -> None:
        super().setUp()
        # Os.kill is used to pause the worker and we do not want to pause tests
        self._init_mock_os_kill()

    def _get_payload_from_worker_com(self, worker_com):
        message = worker_com.get()
        self.assertIsInstance(message, Message)
        return message.payload

    def _assert_task_executed(self, worker_com):
        payload = self._get_payload_from_worker_com(worker_com)
        self.assertIsInstance(payload, TaskExecuted)
        self.assertEqual(WORKER_PID, payload.worker_pid)

    @mock.patch("pcs.daemon.async_tasks.worker.executor.worker_com", Queue())
    def test_successful_run(self, mock_getpid):
        mock_getpid.return_value = WORKER_PID
        executor.task_executor(
            WorkerCommand(
                TASK_IDENT,
                Command(CommandDto("success", {}, COMMAND_OPTIONS)),
                AUTH_USER,
            )
        )
        # 1. TaskExecuted
        self._assert_task_executed(executor.worker_com)
        # 2. TaskFinished
        payload = self._get_payload_from_worker_com(executor.worker_com)
        self.assertIsInstance(payload, TaskFinished)
        self.assertEqual(types.TaskFinishType.SUCCESS, payload.task_finish_type)
        self.assertEqual(RESULT, payload.result)

    @mock.patch("pcs.daemon.async_tasks.worker.executor.worker_com", Queue())
    def test_unsuccessful_run(self, mock_getpid):
        mock_getpid.return_value = WORKER_PID
        executor.task_executor(
            WorkerCommand(
                TASK_IDENT,
                Command(CommandDto("lib_exc", {}, COMMAND_OPTIONS)),
                AUTH_USER,
            )
        )
        # 1. TaskExecuted
        self._assert_task_executed(executor.worker_com)
        # 2. TaskFinished
        payload = self._get_payload_from_worker_com(executor.worker_com)
        self.assertIsInstance(payload, TaskFinished)
        self.assertEqual(types.TaskFinishType.FAIL, payload.task_finish_type)
        self.assertIsNone(payload.result)

    @mock.patch("pcs.daemon.async_tasks.worker.executor.worker_com", Queue())
    def test_unsuccessful_run_additional_reports(self, mock_getpid):
        mock_getpid.return_value = WORKER_PID
        executor.task_executor(
            WorkerCommand(
                TASK_IDENT,
                Command(
                    CommandDto(
                        "lib_exc_reports",
                        {},
                        COMMAND_OPTIONS,
                    )
                ),
                AUTH_USER,
            )
        )
        # 1. TaskExecuted
        self._assert_task_executed(executor.worker_com)
        # 2. Report from the LibraryError exception
        payload = self._get_payload_from_worker_com(executor.worker_com)
        self.assertIsInstance(payload, reports.ReportItemDto)
        # 3. TaskFinished
        payload = self._get_payload_from_worker_com(executor.worker_com)
        self.assertIsInstance(payload, TaskFinished)
        self.assertEqual(types.TaskFinishType.FAIL, payload.task_finish_type)
        self.assertIsNone(payload.result)

    @mock.patch("pcs.daemon.async_tasks.worker.executor.worker_com", Queue())
    def test_unhandled_exception(self, mock_getpid):
        mock_getpid.return_value = WORKER_PID
        executor.task_executor(
            WorkerCommand(
                TASK_IDENT,
                Command(CommandDto("unhandled_exc", {}, COMMAND_OPTIONS)),
                AUTH_USER,
            )
        )
        # 1. TaskExecuted
        self._assert_task_executed(executor.worker_com)
        # 2. TaskFinished
        payload = self._get_payload_from_worker_com(executor.worker_com)
        self.assertIsInstance(payload, TaskFinished)
        self.assertEqual(
            types.TaskFinishType.UNHANDLED_EXCEPTION, payload.task_finish_type
        )
        self.assertIsNone(payload.result)

    @mock.patch("pcs.daemon.async_tasks.worker.executor.worker_com", Queue())
    def test_api_v1_compatibility_disabled(self, mock_getpid):
        mock_getpid.return_value = WORKER_PID
        executor.task_executor(
            WorkerCommand(
                TASK_IDENT,
                Command(CommandDto("success_api_v1", {}, COMMAND_OPTIONS)),
                AUTH_USER,
            )
        )
        # 1. TaskExecuted
        self._assert_task_executed(executor.worker_com)
        # 2. Report from the LibraryError exception
        payload = self._get_payload_from_worker_com(executor.worker_com)
        self.assertIsInstance(payload, reports.ReportItemDto)
        self.assertEqual(payload.message.code, reports.codes.COMMAND_UNKNOWN)
        # 3. TaskFinished
        payload = self._get_payload_from_worker_com(executor.worker_com)
        self.assertIsInstance(payload, TaskFinished)
        self.assertEqual(types.TaskFinishType.FAIL, payload.task_finish_type)

    @mock.patch("pcs.daemon.async_tasks.worker.executor.worker_com", Queue())
    def test_api_v1_compatibility_enabled(self, mock_getpid):
        mock_getpid.return_value = WORKER_PID
        executor.task_executor(
            WorkerCommand(
                TASK_IDENT,
                Command(
                    CommandDto("success_api_v1", {}, COMMAND_OPTIONS),
                    api_v1_compatible=True,
                ),
                AUTH_USER,
            )
        )
        # 1. TaskExecuted
        self._assert_task_executed(executor.worker_com)
        # 2. TaskFinished
        payload = self._get_payload_from_worker_com(executor.worker_com)
        self.assertIsInstance(payload, TaskFinished)
        self.assertEqual(types.TaskFinishType.SUCCESS, payload.task_finish_type)
        self.assertEqual(RESULT, payload.result)



from contextlib import contextmanager
from unittest import TestCase

from pcs.daemon import session
from pcs.daemon.session import Session

from pcs_test.tools.misc import create_setup_patch_mixin

SID = "abc"
USER = "user"
GROUPS = ["group1", "group2"]


PatchSessionMixin = create_setup_patch_mixin(session)


class SessionTest(TestCase, PatchSessionMixin):
    def setUp(self):
        self.now = self.setup_patch("now", return_value=0)
        self.session = Session(SID, USER)

    def test_session_grows_older(self):
        self.now.return_value = 10.1
        self.assertTrue(self.session.was_unused_last(10))
        self.assertFalse(self.session.was_unused_last(11))

    @contextmanager
    def refresh_test(self):
        self.now.return_value = 10.1
        yield self.session
        self.now.return_value = 11.2
        self.assertTrue(self.session.was_unused_last(1))
        self.assertFalse(self.session.was_unused_last(2))

    def test_session_is_refreshable(self):
        # pylint: disable=pointless-statement
        with self.refresh_test() as session1:
            session1.refresh()
        with self.refresh_test() as session1:
            session1.username
        with self.refresh_test() as session1:
            session1.sid


class StorageTest(TestCase, PatchSessionMixin):
    def setUp(self):
        self.now = self.setup_patch("now", return_value=0)
        self.storage = session.Storage(lifetime_seconds=10)

    def test_does_not_accept_foreign_sid(self):
        self.assertIsNone(self.storage.get("unknown_sid"))

    def test_provides_the_same_session_for_same_sid(self):
        session1 = self.storage.login(USER)
        session2 = self.storage.get(session1.sid)
        self.assertIs(session1, session2)

    def test_can_destroy_session(self):
        session1 = self.storage.login(USER)
        self.storage.destroy(session1.sid)
        self.assertIsNone(self.storage.get(session1.sid))

    def test_can_drop_expired_sessions_explicitly(self):
        session1 = self.storage.login(USER)
        self.now.return_value = 5
        session2 = self.storage.login(USER)
        self.now.return_value = 12
        self.storage.drop_expired()
        self.assertIsNone(self.storage.get(session1.sid))
        self.assertIs(self.storage.get(session2.sid), session2)

    def test_can_drop_expired_session_implicitly(self):
        session1 = self.storage.login(USER)
        self.now.return_value = 11
        self.storage.login(USER)
        self.assertIsNone(self.storage.get(session1.sid))

    def test_can_login_new_session(self):
        session1 = self.storage.login(USER)
        self.assertIsNotNone(session1)
        self.assertEqual(session1.username, USER)



import logging
from contextlib import contextmanager
from typing import (
    Iterator,
    Optional,
    cast,
)

from pcs.common.file import RawFileError
from pcs.lib.file.instance import FileInstance
from pcs.lib.file.json import JsonParserException
from pcs.lib.interface.config import ParserErrorException

from . import const
from .config.facade import Facade
from .config.parser import ParserError
from .pam import authenticate_user
from .tools import (
    UserGroupsError,
    get_user_groups,
)
from .types import AuthUser


class _UpdateFacadeError(Exception):
    pass


class AuthProvider:
    def __init__(self, logger: logging.Logger) -> None:
        self._logger = logger
        self._config_file_instance = FileInstance.for_pcs_users_config()

    def _get_facade(self) -> Facade:
        try:
            if not self._config_file_instance.raw_file.exists():
                return Facade([])
            return cast(Facade, self._config_file_instance.read_to_facade())
        except ParserError as e:
            self._logger.error(
                "Unable to parse file '%s': %s",
                self._config_file_instance.raw_file.metadata.path,
                e.msg,
            )
        except JsonParserException:
            self._logger.error(
                "Unable to parse file '%s': not valid json",
                self._config_file_instance.raw_file.metadata.path,
            )
        except ParserErrorException:
            self._logger.error(
                "Unable to parse file '%s'",
                self._config_file_instance.raw_file.metadata.path,
            )
        except RawFileError as e:
            self._logger.error(
                "Unable to read file '%s': %s",
                self._config_file_instance.raw_file.metadata.path,
                e.reason,
            )
        return Facade([])

    @contextmanager
    def _update_facade(self) -> Iterator[Facade]:
        try:
            with self._config_file_instance.raw_file.update() as io_buffer:
                content = io_buffer.getvalue()
                facade = Facade([])
                if content:
                    try:
                        facade = cast(
                            Facade,
                            self._config_file_instance.raw_to_facade(content),
                        )
                    except ParserError as e:
                        self._logger.error(
                            "Unable to parse file '%s': %s",
                            self._config_file_instance.raw_file.metadata.path,
                            e.msg,
                        )
                    except JsonParserException:
                        self._logger.error(
                            "Unable to parse file '%s': not valid json",
                            self._config_file_instance.raw_file.metadata.path,
                        )
                    except ParserErrorException:
                        self._logger.error(
                            "Unable to parse file '%s'",
                            self._config_file_instance.raw_file.metadata.path,
                        )
                yield facade
                io_buffer.seek(0)
                io_buffer.truncate()
                io_buffer.write(
                    self._config_file_instance.facade_to_raw(facade)
                )

        except RawFileError as e:
            self._logger.error(
                "Unable to update file '%s': %s",
                self._config_file_instance.raw_file.metadata.path,
                e.reason,
            )
            raise _UpdateFacadeError() from e

    def login_user(self, username: str) -> Optional[AuthUser]:
        try:
            groups = get_user_groups(username)
        except UserGroupsError:
            self._logger.error(
                "Unable to determine groups of user '%s'", username
            )
            return None
        if const.ADMIN_GROUP not in groups:
            self._logger.debug(
                "User '%s' is not a member of '%s' group",
                username,
                const.ADMIN_GROUP,
            )
            return None
        self._logger.debug("Successful login by '%s'", username)
        return AuthUser(username=username, groups=tuple(groups))

    def auth_by_token(self, token: str) -> Optional[AuthUser]:
        username = self._get_facade().get_user(token)
        if username is None:
            return None
        return self.login_user(username)

    def auth_by_username_password(
        self, username: str, password: str
    ) -> Optional[AuthUser]:
        if authenticate_user(username, password):
            return self.login_user(username)
        self._logger.info(
            "Failed login by '%s': bad username or password", username
        )
        return None

    def create_token(self, username: str) -> Optional[str]:
        try:
            with self._update_facade() as facade:
                return facade.add_user(username)
        except _UpdateFacadeError:
            return None



from typing import Optional

from pcs import settings
from pcs.lib.booth.config_parser import ConfigItem
from pcs.lib.booth.constants import AUTHFILE_FIX_OPTION
from pcs.lib.interface.config import FacadeInterface


class ConfigFacade(FacadeInterface):
    @classmethod
    def create(cls, site_list, arbitrator_list):
        """
        Create a minimal config

        iterable site_list -- list of booth sites' addresses
        iterable arbitrator_list -- list of arbitrators' addresses
        """
        return cls(
            [ConfigItem("site", site) for site in site_list]
            + [ConfigItem("arbitrator", arbit) for arbit in arbitrator_list]
        )

    ### peers

    def get_arbitrators(self):
        return self.__pick_values_by_key("arbitrator")

    def get_sites(self):
        return self.__pick_values_by_key("site")

    ### tickets

    def add_ticket(self, ticket_name, ticket_options):
        """
        Add a booth ticket to the booth config

        string ticket_name -- the name of the ticket
        dict ticket_options -- ticket options
        """
        self._config.append(
            ConfigItem(
                "ticket",
                ticket_name,
                [
                    ConfigItem(key, value)
                    for key, value in sorted(ticket_options.items())
                ],
            )
        )

    def has_ticket(self, ticket_name):
        """
        Return True if the ticket exists, False otherwise

        string ticket_name -- the name of the ticket
        """
        for key, value, dummy_details in self._config:
            if key == "ticket" and value == ticket_name:
                return True
        return False

    def remove_ticket(self, ticket_name):
        """
        Remove an existing booth ticket from the booth config

        string ticket_name -- the name of the ticket
        """
        self._config = [
            config_item
            for config_item in self._config
            if config_item.key != "ticket" or config_item.value != ticket_name
        ]

    ### authfile

    def set_authfile(self, auth_file):
        """
        Set the path to a booth authfile to the booth config

        string auth_file -- the path to a booth authfile
        """
        if settings.booth_enable_authfile_set_enabled:
            self.set_option(AUTHFILE_FIX_OPTION, "yes")
        self.set_option("authfile", auth_file)

    def get_authfile(self) -> Optional[str]:
        """
        Get the path to a booth authfile set in the booth config or None
        """
        return self.get_option("authfile")

    ### tools

    def set_option(self, key: str, value: str) -> None:
        self._config = self.__filter_out_by_key(key)
        if value:
            self._config.insert(0, ConfigItem(key, value))

    def get_option(self, option: str) -> Optional[str]:
        for key, value, dummy_details in reversed(self._config):
            if key == option:
                return value
        return None

    def __pick_values_by_key(self, key):
        return [item.value for item in self._config if item.key == key]

    def __filter_out_by_key(self, key):
        return [item for item in self._config if item.key != key]



import signal
import subprocess
from logging import Logger
from shlex import quote as shell_quote
from typing import (
    Dict,
    Mapping,
    Optional,
    Tuple,
)

from pcs import settings
from pcs.common import reports
from pcs.common.reports import ReportProcessor
from pcs.common.reports.item import ReportItem
from pcs.common.str_tools import join_multilines
from pcs.common.types import StringSequence
from pcs.lib.errors import LibraryError


class KillServicesError(Exception):
    # pylint: disable=super-init-not-called
    def __init__(self, service, message=None, instance=None):
        self.service = service
        self.message = message
        self.instance = instance


class CommandRunner:
    def __init__(
        self,
        logger: Logger,
        reporter: ReportProcessor,
        env_vars: Optional[Mapping[str, str]] = None,
    ):
        self._logger = logger
        self._reporter = reporter
        # Reset environment variables by empty dict is desired here.  We need
        # to get rid of defaults - we do not know the context and environment
        # where the library runs.  We also get rid of PATH settings, so all
        # executables must be specified with full path unless the PATH variable
        # is set from outside.
        self._env_vars = env_vars if env_vars else {}

    @property
    def env_vars(self) -> Dict[str, str]:
        return dict(self._env_vars)

    def run(
        self,
        args: StringSequence,
        stdin_string: Optional[str] = None,
        env_extend: Optional[Mapping[str, str]] = None,
        binary_output: bool = False,
    ) -> Tuple[str, str, int]:
        # Allow overriding default settings. If a piece of code really wants to
        # set own PATH or CIB_file, we must allow it. I.e. it wants to run
        # a pacemaker tool on a CIB in a file but cannot afford the risk of
        # changing the CIB in the file specified by the user.
        env_vars = dict(self._env_vars)
        env_vars.update(dict(env_extend) if env_extend else {})

        log_args = " ".join([shell_quote(x) for x in args])
        self._logger.debug(
            "Running: {args}\nEnvironment:{env_vars}{stdin_string}".format(
                args=log_args,
                stdin_string=(
                    ""
                    if not stdin_string
                    else (
                        "\n--Debug Input Start--\n{0}\n--Debug Input End--"
                    ).format(stdin_string)
                ),
                env_vars=(
                    ""
                    if not env_vars
                    else (
                        "\n"
                        + "\n".join(
                            [
                                "  {0}={1}".format(key, val)
                                for key, val in sorted(env_vars.items())
                            ]
                        )
                    )
                ),
            )
        )
        self._reporter.report(
            ReportItem.debug(
                reports.messages.RunExternalProcessStarted(
                    log_args,
                    stdin_string,
                    env_vars,
                )
            )
        )

        try:
            # pylint: disable=subprocess-popen-preexec-fn, consider-using-with
            # this is OK as pcs is only single-threaded application
            process = subprocess.Popen(
                args,
                # Some commands react differently if they get anything via stdin
                stdin=(
                    subprocess.PIPE
                    if stdin_string is not None
                    else subprocess.DEVNULL
                ),
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                preexec_fn=(
                    lambda: signal.signal(signal.SIGPIPE, signal.SIG_DFL)
                ),
                close_fds=True,
                shell=False,
                env=env_vars,
                # decodes newlines and in python3 also converts bytes to str
                universal_newlines=(not binary_output),
            )
            out_std, out_err = process.communicate(stdin_string)
            retval = process.returncode
        except OSError as e:
            raise LibraryError(
                ReportItem.error(
                    reports.messages.RunExternalProcessError(
                        log_args,
                        e.strerror,
                    )
                )
            ) from e

        self._logger.debug(
            (
                "Finished running: {args}\nReturn value: {retval}"
                + "\n--Debug Stdout Start--\n{out_std}\n--Debug Stdout End--"
                + "\n--Debug Stderr Start--\n{out_err}\n--Debug Stderr End--"
            ).format(
                args=log_args, retval=retval, out_std=out_std, out_err=out_err
            )
        )
        self._reporter.report(
            ReportItem.debug(
                reports.messages.RunExternalProcessFinished(
                    log_args,
                    retval,
                    out_std,
                    out_err,
                )
            )
        )
        return out_std, out_err, retval


def kill_services(runner, services):
    """
    Kill specified services in local system
    CommandRunner runner
    iterable services service names
    """
    # make killall not report that a process is not running
    stdout, stderr, retval = runner.run(
        [settings.killall_executable, "--quiet", "--signal", "9", "--"]
        + list(services)
    )
    # If a process isn't running, killall will still return 1 even with --quiet.
    # We don't consider that an error, so we check for output string as well.
    # If it's empty, no actual error happened.
    if retval != 0:
        message = join_multilines([stderr, stdout])
        if message:
            raise KillServicesError(list(services), message)


def is_proxy_set(env_dict):
    """
    Returns True whenever any of proxy environment variables (https_proxy,
    HTTPS_PROXY, all_proxy, ALL_PROXY) are set in env_dict. False otherwise.

    env_dict -- environment variables in dict
    """
    proxy_list = ["https_proxy", "all_proxy"]
    for var in proxy_list + [v.upper() for v in proxy_list]:
        if env_dict.get(var, "") != "":
            return True
    return False



from functools import partial

from pcs import (
    resource,
    usage,
)
from pcs.cli.common.routing import create_router
from pcs.cli.resource.relations import show_resource_relations_cmd

from .resource_stonith_common import (
    resource_defaults_cmd,
    resource_op_defaults_cmd,
)

resource_cmd = create_router(
    {
        "help": lambda lib, argv, modifiers: print(usage.resource(argv)),
        "list": resource.resource_list_available,
        "describe": resource.resource_list_options,
        "create": resource.resource_create,
        "move": resource.resource_move,
        "move-with-constraint": resource.resource_move_with_constraint,
        "ban": resource.resource_ban,
        "clear": resource.resource_unmove_unban,
        "standards": resource.resource_standards,
        "providers": resource.resource_providers,
        "agents": resource.resource_agents,
        "update": resource.update_cmd,
        "meta": resource.meta_cmd,
        "delete": resource.resource_remove_cmd,
        "remove": resource.resource_remove_cmd,
        # TODO remove, deprecated command
        # replaced with 'resource status' and 'resource config'
        "show": resource.resource_show,
        "status": resource.resource_status,
        "config": resource.config,
        "group": create_router(
            {
                "add": resource.resource_group_add_cmd,
                "list": resource.resource_group_list,
                "remove": resource.resource_group_rm_cmd,
                "delete": resource.resource_group_rm_cmd,
            },
            ["resource", "group"],
        ),
        "ungroup": resource.resource_group_rm_cmd,
        "clone": resource.resource_clone,
        "promotable": partial(resource.resource_clone, promotable=True),
        "unclone": resource.resource_clone_master_remove,
        "enable": resource.resource_enable_cmd,
        "disable": resource.resource_disable_cmd,
        "safe-disable": resource.resource_safe_disable_cmd,
        "restart": resource.resource_restart,
        "debug-start": partial(
            resource.resource_force_action, action="debug-start"
        ),
        "debug-stop": partial(
            resource.resource_force_action, action="debug-stop"
        ),
        "debug-promote": partial(
            resource.resource_force_action, action="debug-promote"
        ),
        "debug-demote": partial(
            resource.resource_force_action, action="debug-demote"
        ),
        "debug-monitor": partial(
            resource.resource_force_action, action="debug-monitor"
        ),
        "manage": resource.resource_manage_cmd,
        "unmanage": resource.resource_unmanage_cmd,
        "failcount": create_router(
            {
                "show": resource.resource_failcount_show,
            },
            ["resource", "failcount"],
            default_cmd="show",
        ),
        "op": create_router(
            {
                "defaults": resource_op_defaults_cmd(
                    ["resource", "op", "defaults"]
                ),
                "add": resource.op_add_cmd,
                "remove": resource.op_delete_cmd,
                "delete": resource.op_delete_cmd,
            },
            ["resource", "op"],
        ),
        "defaults": resource_defaults_cmd(["resource", "defaults"]),
        "cleanup": resource.resource_cleanup,
        "refresh": resource.resource_refresh,
        "relocate": create_router(
            {
                "show": resource.resource_relocate_show_cmd,
                "dry-run": resource.resource_relocate_dry_run_cmd,
                "run": resource.resource_relocate_run_cmd,
                "clear": resource.resource_relocate_clear_cmd,
            },
            ["resource", "relocate"],
        ),
        "utilization": resource.resource_utilization_cmd,
        "bundle": create_router(
            {
                "create": resource.resource_bundle_create_cmd,
                "reset": resource.resource_bundle_reset_cmd,
                "update": resource.resource_bundle_update_cmd,
            },
            ["resource", "bundle"],
        ),
        # internal use only
        "get_resource_agent_info": resource.get_resource_agent_info,
        "relations": show_resource_relations_cmd,
    },
    ["resource"],
    default_cmd="status",
)



from dataclasses import dataclass
from typing import Optional

from pcs.common.tools import StringCollection

from . import const


@dataclass(frozen=True)
class AuthUser:
    username: str
    groups: StringCollection

    @property
    def is_superuser(self) -> bool:
        return self.username == const.SUPERUSER


@dataclass(frozen=True)
class DesiredUser:
    username: Optional[str]
    groups: Optional[StringCollection]
